# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2015-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import sys
from unittest import mock

from testtools.matchers import Equals

from snapcraft.internal import sources
from tests import unit


class TestDeb(unit.FakeFileHTTPServerBasedTestCase):
    def setUp(self):
        super().setUp()

        patcher = mock.patch("debian.arfile.ArFile")
        self.mock_ar = patcher.start()
        self.mock_ar.return_value.getnames.return_value = [
            "data.tar.gz",
            "control.tar.gz",
            "debian-binary",
        ]
        self.addCleanup(patcher.stop)

        patcher = mock.patch("tarfile.open")
        patcher.start()
        self.addCleanup(patcher.stop)

    def test_pull_debfile_must_download_and_extract(self):
        dest_dir = "src"
        os.makedirs(dest_dir)
        deb_file_name = "test.deb"
        source = "http://{}:{}/{file_name}".format(
            *self.server.server_address, file_name=deb_file_name
        )
        deb_source = sources.Deb(source, dest_dir)

        deb_source.pull()

        self.mock_ar.assert_called_once_with(
            os.path.join(deb_source.source_dir, deb_file_name)
        )

    def test_extract_and_keep_debfile(self):
        deb_file_name = "test.deb"
        source = "http://{}:{}/{file_name}".format(
            *self.server.server_address, file_name=deb_file_name
        )
        dest_dir = os.path.abspath(os.curdir)
        deb_source = sources.Deb(source, dest_dir)

        deb_source.download()
        deb_source.provision(dst=dest_dir, keep_deb=True)

        deb_download = os.path.join(deb_source.source_dir, deb_file_name)
        self.mock_ar.assert_called_once_with(
            os.path.join(deb_source.source_dir, deb_file_name)
        )

        with open(deb_download, "r") as deb_file:
            self.assertThat(deb_file.read(), Equals("Test fake file"))

    def test_has_source_handler_entry_on_linux(self):
        if sys.platform == "linux":
            self.assertTrue(sources._source_handler["deb"] is sources.Deb)
        else:
            self.assertRaises(KeyError, sources._source_handler["deb"])

    def test_invalid_deb(self):
        self.mock_ar.return_value.getnames.return_value = [
            "control.tar.gz",
            "debian-binary",
        ]

        deb_file_name = "test.deb"
        source = "http://{}:{}/{file_name}".format(
            *self.server.server_address, file_name=deb_file_name
        )
        dest_dir = os.path.abspath(os.curdir)
        deb_source = sources.Deb(source, dest_dir)

        deb_source.download()

        self.assertRaises(
            sources.errors.InvalidDebError,
            deb_source.provision,
            dst=dest_dir,
            keep_deb=True,
        )



import os
import socket

if hasattr(socket, "AF_UNIX"):
    from ._unix import FakeSnapd  # noqa: F401
try:
    from ._unittests import (  # noqa: F401
        FakeElf,
        FakeExtension,
        FakeMetadataExtractor,
        FakeMultipass,
        FakePlugin,
        FakeProjectOptions,
        FakeSnapCommand,
        FakeSnapcraftctl,
    )
except ImportError as import_error:
    if os.path.exists(os.path.join(os.path.dirname(__file__), "..", "snapcraft")):
        raise import_error

from ._fixtures import (  # noqa: F401
    BzrRepo,
    CleanEnvironment,
    FakeBaseEnvironment,
    FakeParts,
    FakePartsServerRunning,
    FakePartsWiki,
    FakePartsWikiOrigin,
    FakePartsWikiOriginRunning,
    FakePartsWikiRunning,
    FakePartsWikiWithSlashes,
    FakePartsWikiWithSlashesRunning,
    FakeServerRunning,
    FakeSnapcraftIsASnap,
    FakeSSOServerRunning,
    FakeStore,
    FakeStoreAPIServerRunning,
    FakeStoreSearchServerRunning,
    FakeStoreUploadServerRunning,
    FakeTerminal,
    GitRepo,
    HgRepo,
    SharedCache,
    SnapcraftYaml,
    StagingStore,
    SvnRepo,
    TempCWD,
    TempXDG,
    TestStore,
    WithoutSnapInstalled,
)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2015-2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from unittest import mock

from testtools.matchers import Equals

from snapcraft.internal import steps

from . import LifecycleCommandsBaseTestCase


class CleanCommandTestCase(LifecycleCommandsBaseTestCase):
    def test_unprime_fails(self):
        result = self.run_command(["clean", "--unprime"])

        self.assertThat(result.exit_code, Equals(2))

    def test_using_defaults(self):
        result = self.run_command(["clean"])

        self.assertThat(result.exit_code, Equals(0))
        self.fake_get_provider_for.mock.assert_called_once_with("multipass")
        self.fake_lifecycle_clean.mock.assert_called_once_with(
            mock.ANY, tuple(), steps.PRIME
        )
        self.provider_mock.clean_parts.assert_not_called()
        self.provider_class_mock().clean_project.assert_called_once_with()

    def test_with_parts_specified_using_defaults(self):
        result = self.run_command(["clean", "part0", "part1", "part2"])

        self.assertThat(result.exit_code, Equals(0))
        self.fake_get_provider_for.mock.assert_called_once_with("multipass")
        self.fake_lifecycle_clean.mock.assert_not_called()
        self.provider_mock.clean_parts.assert_called_once_with(
            part_names=tuple(["part0", "part1", "part2"])
        )
        self.provider_class_mock().clean_project.assert_not_called()

    def test_using_lxd(self):
        result = self.run_command(["clean", "--use-lxd"])

        self.assertThat(result.exit_code, Equals(0))
        self.fake_get_provider_for.mock.assert_called_once_with("lxd")
        self.fake_lifecycle_clean.mock.assert_called_once_with(
            mock.ANY, tuple(), steps.PRIME
        )
        self.provider_mock.clean_parts.assert_not_called()
        self.provider_class_mock().clean_project.assert_called_once_with()

    def test_with_parts_specified_using_lxd(self):
        result = self.run_command(["clean", "--use-lxd", "part0", "part1", "part2"])

        self.assertThat(result.exit_code, Equals(0))
        self.fake_get_provider_for.mock.assert_called_once_with("lxd")
        self.fake_lifecycle_clean.mock.assert_not_called()
        self.provider_mock.clean_parts.assert_called_once_with(
            part_names=tuple(["part0", "part1", "part2"])
        )
        self.provider_class_mock().clean_project.assert_not_called()

    def test_using_destructive_mode(self):
        result = self.run_command(["clean", "--destructive-mode"])

        self.assertThat(result.exit_code, Equals(0))
        self.fake_get_provider_for.mock.assert_not_called()
        self.fake_lifecycle_clean.mock.assert_called_once_with(mock.ANY, tuple(), None)

    def test_with_parts_specified_using_destructive_mode(self):
        result = self.run_command(
            ["clean", "--destructive-mode", "part0", "part1", "part2"]
        )

        self.assertThat(result.exit_code, Equals(0))
        self.fake_get_provider_for.mock.assert_not_called()
        self.fake_lifecycle_clean.mock.assert_called_once_with(
            mock.ANY, tuple(["part0", "part1", "part2"]), None
        )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2019 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.


import pytest

from snapcraft.internal.meta import errors
from snapcraft.internal.meta.package_repository import (
    PackageRepository,
    PackageRepositoryApt,
    PackageRepositoryAptPpa,
)


def test_apt_name():
    repo = PackageRepositoryApt(
        architectures=["amd64", "i386"],
        components=["main", "multiverse"],
        formats=["deb", "deb-src"],
        key_id="A" * 40,
        key_server="keyserver.ubuntu.com",
        suites=["xenial", "xenial-updates"],
        url="http://archive.ubuntu.com/ubuntu",
    )

    assert repo.name == "http_archive_ubuntu_com_ubuntu"


@pytest.mark.parametrize(
    "arch", ["amd64", "armhf", "arm64", "i386", "ppc64el", "riscv", "s390x"]
)
def test_apt_valid_architectures(arch):
    package_repo = PackageRepositoryApt(
        key_id="A" * 40, url="http://test", architectures=[arch]
    )

    assert package_repo.architectures == [arch]


def test_apt_invalid_url():
    with pytest.raises(errors.PackageRepositoryValidationError) as exc_info:
        PackageRepositoryApt(
            key_id="A" * 40, url="",
        )

    assert exc_info.value.brief == "Invalid URL ''."
    assert exc_info.value.details == "URLs must be non-empty strings."
    assert (
        exc_info.value.resolution
        == "Verify the repository configuration and ensure that 'url' is correctly specified."
    )


def test_apt_invalid_path():
    with pytest.raises(errors.PackageRepositoryValidationError) as exc_info:
        PackageRepositoryApt(
            key_id="A" * 40, path="", url="http://archive.ubuntu.com/ubuntu",
        )

    assert exc_info.value.brief == "Invalid path ''."
    assert exc_info.value.details == "Paths must be non-empty strings."
    assert (
        exc_info.value.resolution
        == "Verify the repository configuration and ensure that 'path' is a non-empty string such as '/'."
    )


def test_apt_invalid_path_with_suites():
    with pytest.raises(errors.PackageRepositoryValidationError) as exc_info:
        PackageRepositoryApt(
            key_id="A" * 40,
            path="/",
            suites=["xenial", "xenial-updates"],
            url="http://archive.ubuntu.com/ubuntu",
        )

    assert (
        exc_info.value.brief
        == "Suites ['xenial', 'xenial-updates'] cannot be combined with path '/'."
    )
    assert exc_info.value.details == "Path and suites are incomptiable options."
    assert (
        exc_info.value.resolution
        == "Verify the repository configuration and remove 'path' or 'suites'."
    )


def test_apt_invalid_path_with_components():
    with pytest.raises(errors.PackageRepositoryValidationError) as exc_info:
        PackageRepositoryApt(
            key_id="A" * 40,
            path="/",
            components=["main"],
            url="http://archive.ubuntu.com/ubuntu",
        )

    assert (
        exc_info.value.brief == "Components ['main'] cannot be combined with path '/'."
    )
    assert exc_info.value.details == "Path and components are incomptiable options."
    assert (
        exc_info.value.resolution
        == "Verify the repository configuration and remove 'path' or 'components'."
    )


def test_apt_invalid_missing_components():
    with pytest.raises(errors.PackageRepositoryValidationError) as exc_info:
        PackageRepositoryApt(
            key_id="A" * 40,
            suites=["xenial", "xenial-updates"],
            url="http://archive.ubuntu.com/ubuntu",
        )

    assert exc_info.value.brief == "No components specified."
    assert exc_info.value.details == "Components are required when using suites."
    assert (
        exc_info.value.resolution
        == "Verify the repository configuration and ensure that 'components' is correctly specified."
    )


def test_apt_invalid_missing_suites():
    with pytest.raises(errors.PackageRepositoryValidationError) as exc_info:
        PackageRepositoryApt(
            key_id="A" * 40,
            components=["main"],
            url="http://archive.ubuntu.com/ubuntu",
        )

    assert exc_info.value.brief == "No suites specified."
    assert exc_info.value.details == "Suites are required when using components."
    assert (
        exc_info.value.resolution
        == "Verify the repository configuration and ensure that 'suites' is correctly specified."
    )


def test_apt_invalid_suites_as_path():
    with pytest.raises(errors.PackageRepositoryValidationError) as exc_info:
        PackageRepositoryApt(
            key_id="A" * 40,
            suites=["my-suite/"],
            url="http://archive.ubuntu.com/ubuntu",
        )

    assert exc_info.value.brief == "Invalid suite 'my-suite/'."
    assert exc_info.value.details == "Suites must not end with a '/'."
    assert (
        exc_info.value.resolution
        == "Verify the repository configuration and remove the trailing '/ from suites or use the 'path' property to define a path."
    )


def test_apt_marshal():
    repo = PackageRepositoryApt(
        architectures=["amd64", "i386"],
        components=["main", "multiverse"],
        formats=["deb", "deb-src"],
        key_id="A" * 40,
        key_server="xkeyserver.ubuntu.com",
        name="test-name",
        suites=["xenial", "xenial-updates"],
        url="http://archive.ubuntu.com/ubuntu",
    )

    assert repo.marshal() == {
        "architectures": ["amd64", "i386"],
        "components": ["main", "multiverse"],
        "formats": ["deb", "deb-src"],
        "key-id": "A" * 40,
        "key-server": "xkeyserver.ubuntu.com",
        "name": "test-name",
        "suites": ["xenial", "xenial-updates"],
        "type": "apt",
        "url": "http://archive.ubuntu.com/ubuntu",
    }


def test_apt_unmarshal_invalid_extra_keys():
    test_dict = {
        "architectures": ["amd64", "i386"],
        "components": ["main", "multiverse"],
        "formats": ["deb", "deb-src"],
        "key-id": "A" * 40,
        "key-server": "keyserver.ubuntu.com",
        "name": "test-name",
        "suites": ["xenial", "xenial-updates"],
        "type": "apt",
        "url": "http://archive.ubuntu.com/ubuntu",
        "foo": "bar",
        "foo2": "bar",
    }

    with pytest.raises(errors.PackageRepositoryValidationError) as exc_info:
        PackageRepositoryApt.unmarshal(test_dict)

    assert (
        exc_info.value.brief
        == "Found unsupported package repository properties 'foo', 'foo2'."
    )
    assert exc_info.value.details is None
    assert (
        exc_info.value.resolution
        == "Verify repository configuration and ensure it is correct."
    )


def test_apt_unmarshal_invalid_data():
    test_dict = "not-a-dict"

    with pytest.raises(errors.PackageRepositoryValidationError) as exc_info:
        PackageRepositoryApt.unmarshal(test_dict)

    assert exc_info.value.brief == "Invalid package repository object 'not-a-dict'."
    assert (
        exc_info.value.details
        == "Package repository must be a valid dictionary object."
    )
    assert (
        exc_info.value.resolution
        == "Verify repository configuration and ensure that the correct syntax is used."
    )


def test_apt_unmarshal_invalid_type():
    test_dict = {
        "architectures": ["amd64", "i386"],
        "components": ["main", "multiverse"],
        "formats": ["deb", "deb-src"],
        "key-id": "A" * 40,
        "key-server": "keyserver.ubuntu.com",
        "name": "test-name",
        "suites": ["xenial", "xenial-updates"],
        "type": "aptx",
        "url": "http://archive.ubuntu.com/ubuntu",
    }

    with pytest.raises(errors.PackageRepositoryValidationError) as exc_info:
        PackageRepositoryApt.unmarshal(test_dict)

    assert exc_info.value.brief == "Unsupported type 'aptx'."
    assert exc_info.value.details == "The only currently supported type is 'apt'."
    assert (
        exc_info.value.resolution
        == "Verify repository configuration and ensure that 'type' is correctly specified."
    )


def test_ppa_marshal():
    repo = PackageRepositoryAptPpa(ppa="test/ppa")

    assert repo.marshal() == {"type": "apt", "ppa": "test/ppa"}


def test_ppa_invalid_ppa():
    with pytest.raises(errors.PackageRepositoryValidationError) as exc_info:
        PackageRepositoryAptPpa(ppa="")

    assert exc_info.value.brief == "Invalid PPA ''."
    assert exc_info.value.details == "PPAs must be non-empty strings."
    assert (
        exc_info.value.resolution
        == "Verify repository configuration and ensure that 'ppa' is correctly specified."
    )


def test_ppa_unmarshal_invalid_data():
    test_dict = "not-a-dict"

    with pytest.raises(errors.PackageRepositoryValidationError) as exc_info:
        PackageRepositoryAptPpa.unmarshal(test_dict)

    assert exc_info.value.brief == "Invalid package repository object 'not-a-dict'."
    assert (
        exc_info.value.details
        == "Package repository must be a valid dictionary object."
    )
    assert (
        exc_info.value.resolution
        == "Verify repository configuration and ensure that the correct syntax is used."
    )


def test_ppa_unmarshal_invalid_apt_ppa_type():
    test_dict = {"type": "aptx", "ppa": "test/ppa"}

    with pytest.raises(errors.PackageRepositoryValidationError) as exc_info:
        PackageRepositoryAptPpa.unmarshal(test_dict)

    assert exc_info.value.brief == "Unsupported type 'aptx'."
    assert exc_info.value.details == "The only currently supported type is 'apt'."
    assert (
        exc_info.value.resolution
        == "Verify repository configuration and ensure that 'type' is correctly specified."
    )


def test_ppa_unmarshal_invalid_apt_ppa_extra_keys():
    test_dict = {"type": "apt", "ppa": "test/ppa", "test": "foo"}

    with pytest.raises(errors.PackageRepositoryValidationError) as exc_info:
        PackageRepositoryAptPpa.unmarshal(test_dict)

    assert (
        exc_info.value.brief
        == "Found unsupported package repository properties 'test'."
    )
    assert exc_info.value.details is None
    assert (
        exc_info.value.resolution
        == "Verify repository configuration and ensure that it is correct."
    )


def test_unmarshal_package_repositories_list_none():
    assert PackageRepository.unmarshal_package_repositories(None) == list()


def test_unmarshal_package_repositories_list_empty():
    assert PackageRepository.unmarshal_package_repositories(list()) == list()


def test_unmarshal_package_repositories_list_ppa():
    test_dict = {"type": "apt", "ppa": "test/foo"}
    test_list = [test_dict]

    unmarshalled_list = [
        repo.marshal()
        for repo in PackageRepository.unmarshal_package_repositories(test_list)
    ]

    assert unmarshalled_list == test_list


def test_unmarshal_package_repositories_list_apt():
    test_dict = {
        "architectures": ["amd64", "i386"],
        "components": ["main", "multiverse"],
        "formats": ["deb", "deb-src"],
        "key-id": "A" * 40,
        "key-server": "keyserver.ubuntu.com",
        "name": "test-name",
        "suites": ["xenial", "xenial-updates"],
        "type": "apt",
        "url": "http://archive.ubuntu.com/ubuntu",
    }

    test_list = [test_dict]

    unmarshalled_list = [
        repo.marshal()
        for repo in PackageRepository.unmarshal_package_repositories(test_list)
    ]

    assert unmarshalled_list == test_list


def test_unmarshal_package_repositories_list_all():
    test_ppa = {"type": "apt", "ppa": "test/foo"}

    test_deb = {
        "architectures": ["amd64", "i386"],
        "components": ["main", "multiverse"],
        "formats": ["deb", "deb-src"],
        "key-id": "A" * 40,
        "key-server": "keyserver.ubuntu.com",
        "name": "test-name",
        "suites": ["xenial", "xenial-updates"],
        "type": "apt",
        "url": "http://archive.ubuntu.com/ubuntu",
    }

    test_list = [test_ppa, test_deb]

    unmarshalled_list = [
        repo.marshal()
        for repo in PackageRepository.unmarshal_package_repositories(test_list)
    ]

    assert unmarshalled_list == test_list


def test_unmarshal_package_repositories_invalid_data():
    with pytest.raises(errors.PackageRepositoryValidationError) as exc_info:
        PackageRepository.unmarshal_package_repositories("not-a-list")

    assert (
        exc_info.value.brief == "Invalid package-repositories list object 'not-a-list'."
    )
    assert exc_info.value.details == "Package repositories must be a list of objects."
    assert (
        exc_info.value.resolution
        == "Verify 'package-repositories' configuration and ensure that the correct syntax is used."
    )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import abc
import logging
from datetime import datetime
from typing import Any, ClassVar, Dict

import tinydb

logger = logging.getLogger(__name__)


class Migration(metaclass=abc.ABCMeta):
    """Migrates schema from <SCHEMA_VERSION-1> to <SCHEMA_VERSION>."""

    SCHEMA_VERSION: ClassVar[int] = 0

    def __init__(self, *, db: tinydb.TinyDB, snapcraft_version: str) -> None:
        self.db = db
        self._snapcraft_version = snapcraft_version

    def _query_control_record(self) -> Dict[str, Any]:
        """Query control record (single document in 'control' table)."""
        control_table = self.db.table("control")
        control_records = control_table.all()

        if len(control_records) == 0:
            return dict(schema_version=0)
        elif len(control_records) == 1:
            return control_records[0]

        raise RuntimeError(f"Invalid control records: {control_records!r}")

    def _update_control_schema_version(self) -> None:
        """Update 'control' table record to SCHEMA_VERSION."""
        control_record = self._query_control_record()
        control_record["schema_version"] = self.SCHEMA_VERSION

        control_table = self.db.table("control")
        control_table.truncate()
        control_table.insert(control_record)

    def _record_migration(self) -> None:
        """Record migration in 'migration' table."""
        migration_table = self.db.table("migration")
        migration_table.insert(
            {
                "schema_version": self.SCHEMA_VERSION,
                "snapcraft_version": self._snapcraft_version,
                "timestamp": datetime.utcnow().isoformat() + "Z",
            }
        )

    @abc.abstractmethod
    def _migrate(self) -> None:
        """Per-migration implementation."""
        ...

    def apply(self) -> int:
        """Apply migration, if determined to be necessary.

        Returns current schema version."""
        control_record = self._query_control_record()
        current_schema_version = control_record["schema_version"]

        if self.SCHEMA_VERSION <= current_schema_version:
            logger.debug(
                f"Migration apply: migration {self.SCHEMA_VERSION} already applied, ignoring..."
            )
            return current_schema_version

        logger.debug(
            f"Migration apply: applying migration for {self.SCHEMA_VERSION} for {control_record}"
        )
        self._migrate()
        self._record_migration()
        self._update_control_schema_version()
        return self.SCHEMA_VERSION


class MigrationV1(Migration):
    """Default (Initial) Migration to v1."""

    SCHEMA_VERSION: ClassVar[int] = 1

    def _migrate_control(self) -> None:
        control_table = self.db.table("control")
        control_table.insert(
            {
                "created_with_snapcraft_version": self._snapcraft_version,
                "schema_version": self.SCHEMA_VERSION,
            }
        )

    def _migrate(self) -> None:
        """Per-migration implementation."""
        self._migrate_control()



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2017-2021 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from unittest import mock

import fixtures
from testtools.matchers import Equals, FileContains, FileExists, Is

from snapcraft.internal.repo import errors, snaps
from tests import unit


class SnapPackageCurrentChannelTest(unit.TestCase):
    def assert_channels(self, snap, installed_snaps, expected):
        self.fake_snapd.snaps_result = installed_snaps
        snap_pkg = snaps.SnapPackage(snap)
        self.assertThat(snap_pkg.get_current_channel(), Equals(expected))

    def test_risk(self):
        self.assert_channels(
            snap="fake-snap-stable/stable",
            installed_snaps=[{"name": "fake-snap-stable", "channel": "stable"}],
            expected="latest/stable",
        )

    def test_track_risk(self):
        self.assert_channels(
            snap="fake-snap-stable/latest/stable",
            installed_snaps=[{"name": "fake-snap-stable", "channel": "stable"}],
            expected="latest/stable",
        )

    def test_track_risk_branch(self):
        self.assert_channels(
            snap="fake-snap-branch/candidate/branch",
            installed_snaps=[
                {"name": "fake-snap-branch", "channel": "candidate/branch"}
            ],
            expected="latest/candidate/branch",
        )


class SnapPackageIsInstalledTest(unit.TestCase):
    def assert_installed(self, snap, installed_snaps, expected):
        self.fake_snapd.snaps_result = installed_snaps
        snap_pkg = snaps.SnapPackage(snap)
        self.expectThat(snap_pkg.installed, Is(expected))
        self.expectThat(snaps.SnapPackage.is_snap_installed(snap), Is(expected))

    def test_default(self):
        self.assert_installed(
            snap="fake-snap-stable",
            installed_snaps=[{"name": "fake-snap-stable", "channel": "stable"}],
            expected=True,
        )

    def test_track_risk(self):

        self.assert_installed(
            snap="fake-snap-stable/latest/stable",
            installed_snaps=[{"name": "fake-snap-stable", "channel": "stable"}],
            expected=True,
        )

    def test_default_not_installed(self):
        self.assert_installed(snap="missing-snap", installed_snaps=[], expected=False),

    def test_track_risk_not_installed(self):
        self.assert_installed(
            snap="missing-snap/latest/stable", installed_snaps=[], expected=False
        )


class SnapPackageIsInStoreTest(unit.TestCase):
    def assert_in_store(self, snap, find_result, expected):
        self.fake_snapd.find_result = find_result
        snap_pkg = snaps.SnapPackage(snap)
        self.assertThat(snap_pkg.in_store, Is(expected))

    def test_default(self):
        self.assert_in_store(
            snap="fake-snap", find_result=[{"fake-snap": "dummy"}], expected=True
        )

    def test_track_risk(self):
        self.assert_in_store(
            snap="fake-snap/latest/stable",
            find_result=[{"fake-snap": "dummy"}],
            expected=True,
        )

    def test_default_not_in_store(self):
        self.assert_in_store(snap="missing-snap", find_result=[], expected=False)

    def test_track_risk_not_in_store(self):
        self.assert_in_store(
            snap="missing-snap/latest/stable", find_result=[], expected=False
        )


class SnapPackageIsClassicTest(unit.TestCase):
    def assert_classic(self, snap, find_result, expected):
        self.fake_snapd.find_result = find_result
        snap_pkg = snaps.SnapPackage(snap)
        self.assertThat(snap_pkg.is_classic(), Is(expected))

    def test_classic(self):
        self.assert_classic(
            snap="fake-snap/classic/stable",
            find_result=[
                {
                    "fake-snap": {
                        "channels": {"classic/stable": {"confinement": "classic"}}
                    }
                }
            ],
            expected=True,
        )

    def test_strict(self):
        self.assert_classic(
            snap="fake-snap/strict/stable",
            find_result=[
                {
                    "fake-snap": {
                        "channels": {"strict/stable": {"confinement": "strict"}}
                    }
                }
            ],
            expected=False,
        )

    def test_devmode(self):
        self.assert_classic(
            snap="fake-snap/devmode/stable",
            find_result=[
                {
                    "fake-snap": {
                        "channels": {"devmode/stable": {"confinement": "devmode"}}
                    }
                }
            ],
            expected=False,
        )


class SnapPackageIsValidTest(unit.TestCase):
    def assert_valid(self, snap, find_result, expected):
        self.fake_snapd.find_result = find_result
        snap_pkg = snaps.SnapPackage(snap)
        self.expectThat(snap_pkg.is_valid(), Is(expected))
        self.expectThat(snaps.SnapPackage.is_valid_snap(snap), Is(expected))

    def test_default(self):
        self.assert_valid(
            snap="fake-snap",
            find_result=[
                {
                    "fake-snap": {
                        "channels": {"latest/stable": {"confinement": "strict"}}
                    }
                }
            ],
            expected=True,
        )

    def test_track_risk(self):
        self.assert_valid(
            snap="fake-snap/strict/stable",
            find_result=[
                {
                    "fake-snap": {
                        "channels": {"strict/stable": {"confinement": "strict"}}
                    }
                }
            ],
            expected=True,
        )

    def test_invalid_track(self):
        self.assert_valid(
            snap="fake-snap/non-existent/edge",
            find_result=[
                {
                    "fake-snap": {
                        "channels": {"strict/stable": {"confinement": "strict"}}
                    }
                }
            ],
            expected=False,
        )

    def test_missing_snap(self):
        self.assert_valid(snap="missing-snap", find_result=[], expected=False)

    def test_installed(self):
        snap_pkg = snaps.SnapPackage("fake-snap/strict/stable/branch")
        snap_pkg.get_local_snap_info = lambda: {"channel": "strict/stable/branch"}

        self.assertTrue(snap_pkg.is_valid())

    def test_404(self):
        self.fake_snapd.find_code = 404
        self.assert_valid(snap="missing-snap", find_result=[], expected=False)


class SnapPackageLifecycleTest(unit.TestCase):
    def test_install_classic(self):
        self.fake_snapd.find_result = [
            {"fake-snap": {"channels": {"classic/stable": {"confinement": "classic"}}}}
        ]

        snap_pkg = snaps.SnapPackage("fake-snap/classic/stable")
        snap_pkg.install()
        self.assertThat(
            self.fake_snap_command.calls,
            Equals(
                [
                    ["snap", "whoami"],
                    [
                        "sudo",
                        "snap",
                        "install",
                        "fake-snap",
                        "--channel",
                        "classic/stable",
                        "--classic",
                    ],
                ]
            ),
        )

    def test_install_non_classic(self):
        self.fake_snapd.find_result = [
            {"fake-snap": {"channels": {"strict/stable": {"confinement": "strict"}}}}
        ]

        snap_pkg = snaps.SnapPackage("fake-snap/strict/stable")
        snap_pkg.install()
        self.assertThat(
            self.fake_snap_command.calls,
            Equals(
                [
                    ["snap", "whoami"],
                    [
                        "sudo",
                        "snap",
                        "install",
                        "fake-snap",
                        "--channel",
                        "strict/stable",
                    ],
                ]
            ),
        )

    def test_install_classic_not_on_channel(self):
        self.fake_snapd.find_result = [{"fake-snap": {"channels": {}}}]

        snap_pkg = snaps.SnapPackage("fake-snap/classic/stable")
        snap_pkg.install()
        self.assertThat(
            self.fake_snap_command.calls,
            Equals(
                [
                    ["snap", "whoami"],
                    [
                        "sudo",
                        "snap",
                        "install",
                        "fake-snap",
                        "--channel",
                        "classic/stable",
                    ],
                ]
            ),
        )

    def test_install_branch(self):
        snap_pkg = snaps.SnapPackage("fake-snap/strict/stable/branch")
        snap_pkg.install()
        self.assertThat(
            self.fake_snap_command.calls,
            Equals(
                [
                    ["snap", "whoami"],
                    [
                        "sudo",
                        "snap",
                        "install",
                        "fake-snap",
                        "--channel",
                        "strict/stable/branch",
                    ],
                ]
            ),
        )

    def test_download_from_host(self):
        fake_get_assertion = fixtures.MockPatch(
            "snapcraft.internal.repo.snaps.get_assertion", return_value=b"foo-assert"
        )
        self.useFixture(fake_get_assertion)

        self.fake_snapd.snaps_result = [
            {
                "id": "fake-snap-id",
                "name": "fake-snap",
                "channel": "stable",
                "revision": "10",
            }
        ]

        snap_pkg = snaps.SnapPackage("fake-snap/strict/stable")
        snap_pkg.local_download(
            snap_path="fake-snap.snap", assertion_path="fake-snap.assert"
        )

        self.assertThat("fake-snap.snap", FileExists())
        self.assertThat(
            "fake-snap.assert", FileContains("foo-assert\nfoo-assert\nfoo-assert\n")
        )
        fake_get_assertion.mock.assert_has_calls(
            [
                mock.call(
                    [
                        "account-key",
                        "public-key-sha3-384=BWDEoaqyr25nF5SNCvEv2v7QnM9QsfCc0PBMYD_i2NGSQ32EF2d4D0hqUel3m8ul",
                    ]
                ),
                mock.call(["snap-declaration", "snap-name=fake-snap"]),
                mock.call(
                    ["snap-revision", "snap-revision=10", "snap-id=fake-snap-id"]
                ),
            ]
        )

    def test_download_from_host_dangerous(self):
        fake_get_assertion = fixtures.MockPatch(
            "snapcraft.internal.repo.snaps.get_assertion", return_value=b"foo-assert"
        )
        self.useFixture(fake_get_assertion)
        self.fake_snapd.snaps_result = [
            {
                "id": "fake-snap-id",
                "name": "fake-snap",
                "channel": "stable",
                "revision": "x1",
            }
        ]

        snap_pkg = snaps.SnapPackage("fake-snap/strict/stable")
        snap_pkg.local_download(
            snap_path="fake-snap.snap", assertion_path="fake-snap.assert"
        )

        self.assertThat("fake-snap.snap", FileExists())
        self.assertThat("fake-snap.assert", FileContains(""))
        fake_get_assertion.mock.assert_not_called()

    def test_install_logged_in(self):
        self.fake_snapd.find_result = [
            {"fake-snap": {"channels": {"strict/stable": {"confinement": "strict"}}}}
        ]

        self.fake_snap_command.login("user@email.com")
        snap_pkg = snaps.SnapPackage("fake-snap/strict/stable")
        snap_pkg.install()
        self.assertThat(
            self.fake_snap_command.calls,
            Equals(
                [
                    ["snap", "whoami"],
                    ["snap", "install", "fake-snap", "--channel", "strict/stable"],
                ]
            ),
        )

    def test_install_fails(self):
        self.fake_snap_command.install_success = False
        snap_pkg = snaps.SnapPackage("fake-snap/strict/stable")
        self.assertRaises(errors.SnapInstallError, snap_pkg.install)

    def test_refresh(self):
        self.fake_snapd.find_result = [
            {"fake-snap": {"channels": {"strict/stable": {"confinement": "strict"}}}}
        ]

        snap_pkg = snaps.SnapPackage("fake-snap/strict/stable")
        snap_pkg.refresh()
        self.assertThat(
            self.fake_snap_command.calls,
            Equals(
                [
                    ["snap", "whoami"],
                    [
                        "sudo",
                        "snap",
                        "refresh",
                        "fake-snap",
                        "--channel",
                        "strict/stable",
                    ],
                ]
            ),
        )

    def test_refresh_branch(self):
        snap_pkg = snaps.SnapPackage("fake-snap/strict/stable/branch")
        snap_pkg.refresh()
        self.assertThat(
            self.fake_snap_command.calls,
            Equals(
                [
                    ["snap", "whoami"],
                    [
                        "sudo",
                        "snap",
                        "refresh",
                        "fake-snap",
                        "--channel",
                        "strict/stable/branch",
                    ],
                ]
            ),
        )

    def test_download(self):
        self.fake_snapd.find_result = [
            {"fake-snap": {"channels": {"strict/stable": {"confinement": "strict"}}}}
        ]

        snap_pkg = snaps.SnapPackage("fake-snap")
        snap_pkg.download()
        self.assertThat(
            self.fake_snap_command.calls, Equals([["snap", "download", "fake-snap"]])
        )

    def test_download_channel(self):
        self.fake_snapd.find_result = [
            {"fake-snap": {"channels": {"strict/edge": {"confinement": "strict"}}}}
        ]

        snap_pkg = snaps.SnapPackage("fake-snap/strict/stable")
        snap_pkg.download()
        self.assertThat(
            self.fake_snap_command.calls,
            Equals([["snap", "download", "fake-snap", "--channel", "strict/stable"]]),
        )

    def test_download_classic(self):
        self.fake_snapd.find_result = [
            {"fake-snap": {"channels": {"classic/stable": {"confinement": "classic"}}}}
        ]

        snap_pkg = snaps.SnapPackage("fake-snap")
        snap_pkg.download()
        self.assertThat(
            self.fake_snap_command.calls, Equals([["snap", "download", "fake-snap"]])
        )

    def test_download_snaps(self):
        self.fake_snapd.find_result = [
            {"fake-snap": {"channels": {"latest/stable": {"confinement": "strict"}}}},
            {
                "other-fake-snap": {
                    "channels": {"latest/stable": {"confinement": "strict"}}
                }
            },
        ]

        snaps.download_snaps(
            snaps_list=["fake-snap", "other-fake-snap/latest/stable"],
            directory="fakedir",
        )
        self.assertThat(
            self.fake_snap_command.calls,
            Equals(
                [
                    ["snap", "download", "fake-snap"],
                    [
                        "snap",
                        "download",
                        "other-fake-snap",
                        "--channel",
                        "latest/stable",
                    ],
                ]
            ),
        )

    def test_download_snaps_with_invalid(self):
        self.fake_snapd.find_result = [
            {"fake-snap": {"channels": {"latest/stable": {"confinement": "strict"}}}},
        ]
        self.fake_snap_command.download_side_effect = [True, False]

        self.assertRaises(
            errors.SnapDownloadError,
            snaps.download_snaps,
            snaps_list=["fake-snap", "other-invalid"],
            directory="fakedir",
        )
        self.assertThat(
            self.fake_snap_command.calls,
            Equals(
                [
                    ["snap", "download", "fake-snap"],
                    ["snap", "download", "other-invalid"],
                ]
            ),
        )

    def test_refresh_to_classic(self):
        self.fake_snapd.find_result = [
            {"fake-snap": {"channels": {"classic/stable": {"confinement": "classic"}}}}
        ]

        snap_pkg = snaps.SnapPackage("fake-snap/classic/stable")
        snap_pkg.refresh()
        self.assertThat(
            self.fake_snap_command.calls,
            Equals(
                [
                    ["snap", "whoami"],
                    [
                        "sudo",
                        "snap",
                        "refresh",
                        "fake-snap",
                        "--channel",
                        "classic/stable",
                        "--classic",
                    ],
                ]
            ),
        )

    def test_refresh_not_on_channel(self):
        self.fake_snapd.find_result = [{"fake-snap": {"channels": {}}}]

        snap_pkg = snaps.SnapPackage("fake-snap/classic/stable")
        snap_pkg.refresh()
        self.assertThat(
            self.fake_snap_command.calls,
            Equals(
                [
                    ["snap", "whoami"],
                    [
                        "sudo",
                        "snap",
                        "refresh",
                        "fake-snap",
                        "--channel",
                        "classic/stable",
                    ],
                ]
            ),
        )

    def test_refresh_logged_in(self):
        self.fake_snapd.find_result = [
            {"fake-snap": {"channels": {"strict/stable": {"confinement": "strict"}}}}
        ]

        self.fake_snap_command.login("user@email.com")
        snap_pkg = snaps.SnapPackage("fake-snap/strict/stable")
        snap_pkg.refresh()
        self.assertThat(
            self.fake_snap_command.calls,
            Equals(
                [
                    ["snap", "whoami"],
                    ["snap", "refresh", "fake-snap", "--channel", "strict/stable"],
                ]
            ),
        )

    def test_refresh_fails(self):
        snap_pkg = snaps.SnapPackage("fake-snap/strict/stable")
        self.fake_snap_command.refresh_success = False
        self.assertRaises(errors.SnapRefreshError, snap_pkg.refresh)

    def test_install_snaps_returns_revision(self):
        self.fake_snapd.find_result = [
            {
                "fake-snap": {
                    "channel": "stable",
                    "type": "app",
                    "channels": {"latest/stable": {"confinement": "strict"}},
                }
            }
        ]
        self.fake_snapd.snaps_result = [
            {
                "name": "fake-snap",
                "channel": "stable",
                "revision": "test-fake-snap-revision",
            }
        ]

        installed_snaps = snaps.install_snaps(["fake-snap"])
        self.assertThat(installed_snaps, Equals(["fake-snap=test-fake-snap-revision"]))

    def test_install_snaps_non_stable_base(self):
        self.fake_snapd.find_result = [
            {
                "fake-base-snap": {
                    "channel": "beta",
                    "type": "base",
                    "channels": {"latest/beta": {"confinement": "strict"}},
                }
            }
        ]
        self.fake_snapd.snaps_result = [
            {
                "name": "fake-base-snap",
                "channel": "beta",
                "revision": "test-fake-base-snap-revision",
            }
        ]

        installed_snaps = snaps.install_snaps(["fake-base-snap"])
        self.assertThat(
            installed_snaps, Equals(["fake-base-snap=test-fake-base-snap-revision"])
        )


class InstalledSnapsTestCase(unit.TestCase):
    def test_get_installed_snaps(self):
        self.fake_snapd.snaps_result = [
            {"name": "test-snap-1", "revision": "test-snap-1-revision"},
            {"name": "test-snap-2", "revision": "test-snap-2-revision"},
        ]
        installed_snaps = snaps.get_installed_snaps()
        self.assertThat(
            installed_snaps,
            Equals(
                ["test-snap-1=test-snap-1-revision", "test-snap-2=test-snap-2-revision"]
            ),
        )


class SnapdNotInstalledTestCase(unit.TestCase):
    def setUp(self):
        super().setUp()
        socket_path_patcher = mock.patch(
            "snapcraft.internal.repo.snaps.get_snapd_socket_path_template"
        )
        mock_socket_path = socket_path_patcher.start()
        mock_socket_path.return_value = "http+unix://nonexisting"
        self.addCleanup(socket_path_patcher.stop)

    def test_get_installed_snaps(self):
        installed_snaps = snaps.get_installed_snaps()
        self.assertThat(installed_snaps, Equals([]))



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2019-2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
from collections import OrderedDict
from textwrap import dedent
from unittest import mock

from testtools.matchers import Equals

from snapcraft.internal.meta import errors
from snapcraft.internal.meta.snap import Snap
from snapcraft.internal.meta.system_user import SystemUserScope
from tests import unit


class SnapTests(unit.TestCase):
    def test_empty(self):
        snap_dict = OrderedDict()

        snap = Snap()

        self.assertEqual(snap_dict, snap.to_dict())

    def test_empty_from_dict(self):
        snap_dict = OrderedDict({})

        snap = Snap.from_dict(snap_dict=snap_dict)

        self.assertEqual(snap_dict, snap.to_dict())

    def test_missing_keys(self):
        snap_dict = OrderedDict({"name": "snap-test", "grade": "stable"})

        snap = Snap.from_dict(snap_dict=snap_dict)

        self.assertEqual(snap_dict, snap.to_dict())
        self.assertRaises(errors.MissingSnapcraftYamlKeysError, snap.validate)

    def test_grade_devel_statisfies_required_grade(self):
        self.fake_snapd.snaps_result = [
            {"name": "fake-base", "channel": "edge", "revision": "fake-revision"}
        ]

        snap_dict = OrderedDict(
            {
                "name": "snap-test",
                "base": "fake-base",
                "version": "snap-version",
                "summary": "snap-summary",
                "description": "snap-description",
                "grade": "devel",
            }
        )

        snap = Snap.from_dict(snap_dict=snap_dict)

        snap.validate()

    def test_simple(self):
        snap_dict = OrderedDict(
            {
                "name": "snap-test",
                "version": "snap-version",
                "summary": "snap-summary",
                "description": "snap-description",
                "grade": "stable",
            }
        )

        snap = Snap.from_dict(snap_dict=snap_dict)
        snap.validate()

        self.assertEqual(snap_dict, snap.to_dict())
        self.assertEqual(False, snap.is_passthrough_enabled)
        self.assertEqual(snap_dict["name"], snap.name)
        self.assertEqual(snap_dict["version"], snap.version)
        self.assertEqual(snap_dict["summary"], snap.summary)
        self.assertEqual(snap_dict["description"], snap.description)

    def test_snap_yaml_passthrough(self):
        snap_dict = OrderedDict(
            {
                "name": "snap-test",
                "version": "snap-version",
                "summary": "snap-summary",
                "description": "snap-description",
                "passthrough": {"otherkey": "othervalue"},
                "grade": "stable",
            }
        )

        snap = Snap.from_dict(snap_dict=snap_dict)
        snap.validate()

        transformed_dict = snap_dict.copy()
        passthrough = transformed_dict.pop("passthrough")
        transformed_dict.update(passthrough)

        self.assertEqual(transformed_dict, snap.to_snap_yaml_dict())
        self.assertEqual(True, snap.is_passthrough_enabled)
        self.assertEqual(passthrough, snap.passthrough)
        self.assertEqual(snap_dict["name"], snap.name)
        self.assertEqual(snap_dict["version"], snap.version)
        self.assertEqual(snap_dict["summary"], snap.summary)
        self.assertEqual(snap_dict["description"], snap.description)

    def test_snap_yaml_all_keys(self):
        snap_dict = {
            "name": "snap-test",
            "version": "test-version",
            "summary": "test-summary",
            "description": "test-description",
            "adopt-info": "some-part",
            "apps": {"test-app": {"command": "test-app"}},
            "architectures": ["all"],
            "assumes": ["command-chain"],
            "base": "core",
            "confinement": "strict",
            "compression": "xz",
            "environment": {"TESTING": "1"},
            "epoch": 0,
            "grade": "devel",
            "hooks": {"test-hook": {"command-chain": ["cmd1"], "plugs": ["network"]}},
            "layout": {"/target": {"bind": "$SNAP/foo"}},
            "license": "GPL",
            "package-repositories": [
                {"type": "apt", "ppa": "test/ppa"},
                {
                    "type": "apt",
                    "architectures": ["amd64", "i386"],
                    "components": ["main"],
                    "formats": ["deb"],
                    "key-id": "A" * 40,
                    "key-server": "test-key-server.com",
                    "suites": ["xenial"],
                    "url": "http://archive.ubuntu.com",
                },
            ],
            "passthrough": {"test": "value"},
            "plugs": {"test-plug": {"interface": "some-value"}},
            "slots": {"test-slot": {"interface": "some-value"}},
            "system-usernames": {"snap_daemon": {"scope": "shared"}},
            "title": "test-title",
            "type": "base",
        }

        snap = Snap.from_dict(snap_dict=snap_dict)
        snap.validate()

        expected_dict = snap_dict.copy()
        expected_dict.pop("adopt-info")
        expected_dict.pop("base")
        expected_dict.pop("compression")
        expected_dict.pop("package-repositories")
        expected_dict.update(expected_dict.pop("passthrough"))

        self.assertEqual(expected_dict, snap.to_snap_yaml_dict())
        self.assertEqual(True, snap.is_passthrough_enabled)

    def test_all_keys(self):
        snap_dict = {
            "name": "snap-test",
            "version": "test-version",
            "summary": "test-summary",
            "description": "test-description",
            "adopt-info": "some-part",
            "apps": {"test-app": {"command": "test-app"}},
            "architectures": ["all"],
            "assumes": ["command-chain"],
            "base": "core",
            "confinement": "strict",
            "compression": "lzo",
            "environment": {"TESTING": "1"},
            "epoch": 0,
            "grade": "devel",
            "hooks": {"test-hook": {"command-chain": ["cmd1"], "plugs": ["network"]}},
            "layout": {"/target": {"bind": "$SNAP/foo"}},
            "license": "GPL",
            "passthrough": {"test": "value"},
            "plugs": {"test-plug": {"interface": "some-value"}},
            "slots": {"test-slot": {"interface": "some-value"}},
            "system-usernames": {"snap_daemon": {"scope": "shared"}},
            "title": "test-title",
            "type": "base",
        }

        snap = Snap.from_dict(snap_dict=snap_dict)
        snap.validate()

        self.assertEqual(snap_dict, snap.to_dict())
        self.assertEqual(True, snap.is_passthrough_enabled)

    def test_system_usernames_shortform_scope(self):
        snap_dict = OrderedDict(
            {
                "name": "snap-test",
                "version": "test-version",
                "summary": "test-summary",
                "description": "test-description",
                "system-usernames": {"snap_daemon": "shared", "lxd": "shared"},
            }
        )

        snap = Snap.from_dict(snap_dict=snap_dict)
        snap.validate()

        self.assertThat(
            snap.system_usernames["snap_daemon"].name, Equals("snap_daemon")
        )
        self.assertThat(
            snap.system_usernames["snap_daemon"].scope, Equals(SystemUserScope.SHARED)
        )
        self.assertThat(snap.system_usernames["lxd"].name, Equals("lxd"))
        self.assertThat(
            snap.system_usernames["lxd"].scope, Equals(SystemUserScope.SHARED)
        )

    def test_system_usernames_longform_scope(self):
        snap_dict = OrderedDict(
            {
                "name": "snap-test",
                "version": "test-version",
                "summary": "test-summary",
                "description": "test-description",
                "system-usernames": {
                    "snap_daemon": {"scope": "shared"},
                    "lxd": {"scope": "shared"},
                },
            }
        )

        snap = Snap.from_dict(snap_dict=snap_dict)
        snap.validate()

        self.assertThat(
            snap.system_usernames["snap_daemon"].name, Equals("snap_daemon")
        )
        self.assertThat(
            snap.system_usernames["snap_daemon"].scope, Equals(SystemUserScope.SHARED)
        )
        self.assertThat(snap.system_usernames["lxd"].name, Equals("lxd"))
        self.assertThat(
            snap.system_usernames["lxd"].scope, Equals(SystemUserScope.SHARED)
        )

    def test_is_passthrough_enabled_app(self):
        snap_dict = OrderedDict(
            {
                "name": "snap-test",
                "version": "test-version",
                "summary": "test-summary",
                "description": "test-description",
                "apps": {
                    "test-app": {
                        "command": "test-app",
                        "passthrough": {"some-key": "some-value"},
                    }
                },
            }
        )

        snap = Snap.from_dict(snap_dict=snap_dict)
        snap.validate()

        self.assertEqual(True, snap.is_passthrough_enabled)

    def test_is_passthrough_enabled_hook(self):
        snap_dict = OrderedDict(
            {
                "name": "snap-test",
                "version": "test-version",
                "summary": "test-summary",
                "description": "test-description",
                "hooks": {"test-hook": {"passthrough": {"some-key": "some-value"}}},
            }
        )

        snap = Snap.from_dict(snap_dict=snap_dict)
        snap.validate()

        self.assertEqual(True, snap.is_passthrough_enabled)

    def test_from_file(self):
        snap_yaml = dedent(
            """
            name: test-name
            version: "1.0"
            summary: test-summary
            description: test-description
            base: core18
            architectures:
            - amd64
            assumes:
            - snapd2.39
            confinement: classic
            grade: devel
            apps:
              test-app:
                command: test-command
                completer: test-completer
        """
        )

        meta_path = os.path.join(self.path, "meta")
        os.makedirs(meta_path)

        snap_yaml_path = os.path.join(self.path, "meta", "snap.yaml")
        open(snap_yaml_path, "w").write(snap_yaml)

        snap = Snap.from_file(snap_yaml_path)
        snap.validate()

        self.assertEqual("test-name", snap.name)
        self.assertEqual("1.0", snap.version)
        self.assertEqual("test-summary", snap.summary)
        self.assertEqual("test-description", snap.description)
        self.assertEqual(
            OrderedDict({"command": "test-command", "completer": "test-completer"}),
            snap.apps["test-app"].to_dict(),
        )
        self.assertEqual(["amd64"], snap.architectures)
        self.assertEqual({"snapd2.39"}, snap.assumes)
        self.assertEqual("core18", snap.base)
        self.assertEqual("classic", snap.confinement)
        self.assertEqual("devel", snap.grade)

    def test_to_file(self):
        # Ordering matters for verifying the YAML.
        snap_yaml = dedent(
            """
            name: test-name
            version: '1.0'
            summary: test-summary
            description: test-description
            apps:
              test-app:
                command: test-command
                completer: test-completer
            architectures:
            - amd64
            assumes:
            - snapd2.39
            base: core18
            confinement: classic
            grade: devel
        """
        )

        meta_path = os.path.join(self.path, "meta")
        os.makedirs(meta_path)

        snap_yaml_path = os.path.join(self.path, "meta", "snap.yaml")
        open(snap_yaml_path, "w").write(snap_yaml)

        snap = Snap.from_file(snap_yaml_path)
        snap.validate()

        # Write snap yaml.
        snap.write_snap_yaml(path=snap_yaml_path)

        # Read snap yaml.
        written_snap_yaml = open(snap_yaml_path, "r").read()

        # Compare stripped versions (to remove leading/trailing newlines).
        self.assertEqual(snap_yaml.strip(), written_snap_yaml.strip())

    def test_get_provider_content_directories_no_plugs(self):
        snap = Snap()
        self.assertEqual(set([]), snap.get_provider_content_directories())

    def test_get_provider_content_directories_with_content_plugs(self):
        snap_dict = OrderedDict(
            {
                "name": "snap-test",
                "version": "test-version",
                "summary": "test-summary",
                "description": "test-description",
                "plugs": {
                    "test-plug": {
                        "interface": "content",
                        "content": "content",
                        "target": "target",
                        "default-provider": "gtk-common-themes:gtk-3-themes",
                    }
                },
            }
        )

        meta_snap_yaml = dedent(
            """
            name: test-content-snap-meta-snap-yaml
            version: "1.0"
            summary: test-summary
            description: test-description
            base: core18
            architectures:
            - all
            confinement: strict
            grade: stable
            slots:
              test-slot-name:
                interface: content
                source:
                  read:
                  - $SNAP/dir1
                  - $SNAP/dir2
        """
        )

        snap = Snap.from_dict(snap_dict=snap_dict)
        snap.validate()

        patcher = mock.patch("snapcraft.internal.common.get_installed_snap_path")
        mock_core_path = patcher.start()
        mock_core_path.return_value = self.path
        self.addCleanup(patcher.stop)

        meta_path = os.path.join(self.path, "meta")
        os.makedirs(meta_path)

        snap_yaml_path = os.path.join(meta_path, "snap.yaml")
        open(snap_yaml_path, "w").write(meta_snap_yaml)

        expected_content_dirs = set(
            [os.path.join(self.path, "dir1"), os.path.join(self.path, "dir2")]
        )

        self.assertEqual(expected_content_dirs, snap.get_provider_content_directories())

    def test_ensure_command_chain_assumption(self):
        snap_dict = OrderedDict(
            {
                "name": "snap-test",
                "version": "snap-version",
                "summary": "snap-summary",
                "description": "snap-description",
                "apps": {
                    "test-app": {
                        "command": "test-app",
                        "command-chain": ["test-command-chain"],
                    }
                },
            }
        )

        snap = Snap.from_dict(snap_dict=snap_dict)
        snap._ensure_command_chain_assumption()
        snap.validate()

        self.assertEqual({"command-chain"}, snap.assumes)

    def test_build_base_and_write_snap_yaml_skips_base_core(self):
        snap_dict = OrderedDict(
            {
                "name": "snap-test",
                "version": "snap-version",
                "summary": "snap-summary",
                "description": "snap-description",
                "base": "core",
                "grade": "devel",
            }
        )

        snap = Snap.from_dict(snap_dict=snap_dict)
        snap.validate()

        # Write snap yaml.
        snap_yaml_path = os.path.join(self.path, "snap.yaml")
        snap.write_snap_yaml(path=snap_yaml_path)

        # Read snap yaml.
        written_snap_yaml = open(snap_yaml_path, "r").read()

        self.assertEqual(snap_dict, snap.to_dict())
        self.assertFalse("base" in written_snap_yaml)
        self.assertEqual(snap.get_build_base(), "core")

    def test_build_base_write_snap_yaml_skips_build_base(self):
        snap_dict = OrderedDict(
            {
                "name": "snap-test",
                "version": "snap-version",
                "summary": "snap-summary",
                "description": "snap-description",
                "base": "core20",
                "build-base": "core18",
                "grade": "devel",
            }
        )

        snap = Snap.from_dict(snap_dict=snap_dict)
        snap.validate()

        # Write snap yaml.
        snap_yaml_path = os.path.join(self.path, "snap.yaml")
        snap.write_snap_yaml(path=snap_yaml_path)

        # Read snap yaml.
        written_snap_yaml = open(snap_yaml_path, "r").read()

        self.assertEqual(snap_dict, snap.to_dict())
        self.assertFalse("build-base" in written_snap_yaml)
        self.assertEqual(snap.get_build_base(), "core18")

    def test_build_base_and_write_snap_yaml_with_base_core20(self):
        snap_dict = OrderedDict(
            {
                "name": "snap-test",
                "version": "snap-version",
                "summary": "snap-summary",
                "description": "snap-description",
                "base": "core20",
                "grade": "devel",
            }
        )

        snap = Snap.from_dict(snap_dict=snap_dict)
        snap.validate()

        # Write snap yaml.
        snap_yaml_path = os.path.join(self.path, "snap.yaml")
        snap.write_snap_yaml(path=snap_yaml_path)

        # Read snap yaml.
        written_snap_yaml = open(snap_yaml_path, "r").read()

        self.assertEqual(snap_dict, snap.to_dict())
        self.assertTrue("base" in written_snap_yaml)
        self.assertEqual(snap.get_build_base(), "core20")

    def test_build_base_and_write_snap_yaml_type_base(self):
        snap_dict = OrderedDict(
            {
                "name": "core18",
                "version": "snap-version",
                "summary": "snap-summary",
                "description": "snap-description",
                "grade": "devel",
                "type": "base",
            }
        )

        snap = Snap.from_dict(snap_dict=snap_dict)
        snap.validate()

        # Write snap yaml.
        snap_yaml_path = os.path.join(self.path, "snap.yaml")
        snap.write_snap_yaml(path=snap_yaml_path)

        # Read snap yaml.
        written_snap_yaml = open(snap_yaml_path, "r").read()

        self.assertEqual(snap_dict, snap.to_dict())
        self.assertTrue("base" in written_snap_yaml)
        self.assertEqual(snap.get_build_base(), "core18")

    def test_build_base_and_write_snap_yaml_type_base_with_build_base(self):
        snap_dict = OrderedDict(
            {
                "name": "core20",
                "version": "snap-version",
                "summary": "snap-summary",
                "description": "snap-description",
                "grade": "devel",
                "build-base": "core18",
                "type": "base",
            }
        )

        snap = Snap.from_dict(snap_dict=snap_dict)
        snap.validate()

        # Write snap yaml.
        snap_yaml_path = os.path.join(self.path, "snap.yaml")
        snap.write_snap_yaml(path=snap_yaml_path)

        # Read snap yaml.
        written_snap_yaml = open(snap_yaml_path, "r").read()

        self.assertTrue("base" in written_snap_yaml)
        self.assertFalse("build-base" in written_snap_yaml)
        self.assertEqual(snap.get_build_base(), "core18")


class TestYAMLComparisons:
    scenarios = [
        (
            "slot-all-forms",
            dict(
                snapcraft_yaml=dedent(
                    """
            name: test-name
            version: '1.0'
            summary: test-summary
            description: test-description
            apps:
              test-app:
                command: test-command
            architectures:
            - amd64
            base: core18
            confinement: classic
            grade: devel
            slots:
              long-form:
                interface: content
                content: explicit-content
                read:
                - /
              short-form: interface-name
              shortest-form:
        """
                ),
                snap_yaml=dedent(
                    """
            name: test-name
            version: '1.0'
            summary: test-summary
            description: test-description
            apps:
              test-app:
                command: test-command
            architectures:
            - amd64
            base: core18
            confinement: classic
            grade: devel
            slots:
              long-form:
                interface: content
                content: explicit-content
                read:
                - /
              short-form: interface-name
              shortest-form: null
        """
                ),
            ),
        ),
        (
            "plug-all-forms",
            dict(
                snapcraft_yaml=dedent(
                    """
            name: test-name
            version: '1.0'
            summary: test-summary
            description: test-description
            apps:
              test-app:
                command: test-command
            architectures:
            - amd64
            base: core18
            confinement: classic
            grade: devel
            plugs:
              long-form:
                interface: content
                target: $SNAP/data-dir
                default-provider: some-provider
              short-form: interface-name
              shortest-form:
        """
                ),
                snap_yaml=dedent(
                    """
            name: test-name
            version: '1.0'
            summary: test-summary
            description: test-description
            apps:
              test-app:
                command: test-command
            architectures:
            - amd64
            base: core18
            confinement: classic
            grade: devel
            plugs:
              long-form:
                interface: content
                target: $SNAP/data-dir
                default-provider: some-provider
              short-form: interface-name
              shortest-form: null
        """
                ),
            ),
        ),
    ]

    def test_conversions(self, tmp_work_path, snapcraft_yaml, snap_yaml):
        # Ordering matters for verifying the YAML.
        snapcraft_yaml_path = tmp_work_path / "snapcraft.yaml"
        with snapcraft_yaml_path.open("w") as snapcraft_file:
            print(snapcraft_yaml, file=snapcraft_file)

        snap = Snap.from_file(snapcraft_yaml_path.as_posix())
        snap.validate()

        # Write snap yaml.
        snap_yaml_path = tmp_work_path / "snap.yaml"
        snap.write_snap_yaml(path=snap_yaml_path.as_posix())

        # Read snap yaml.
        with snap_yaml_path.open() as snap_file:
            written_snap_yaml = snap_file.read()

        # Compare stripped versions (to remove leading/trailing newlines).
        assert snap_yaml.strip() == written_snap_yaml.strip()



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2017, 2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import doctest

import pytest

import snapcraft
import snapcraft.internal.project_loader.grammar._try as _try
from snapcraft.internal.project_loader import grammar


def load_tests(loader, tests, ignore):
    tests.addTests(doctest.DocTestSuite(_try))
    return tests


class TestTryStatementGrammar:

    scenarios = [
        (
            "followed body",
            {
                "body": ["foo", "bar"],
                "else_bodies": [],
                "expected_packages": ["foo", "bar"],
            },
        ),
        (
            "followed else",
            {
                "body": ["invalid"],
                "else_bodies": [["valid"]],
                "expected_packages": ["valid"],
            },
        ),
        (
            "optional without else",
            {"body": ["invalid"], "else_bodies": [], "expected_packages": list()},
        ),
        (
            "followed chained else",
            {
                "body": ["invalid1"],
                "else_bodies": [["invalid2"], ["finally-valid"]],
                "expected_packages": ["finally-valid"],
            },
        ),
        (
            "nested body followed body",
            {
                "body": [{"try": ["foo"]}, {"else": ["bar"]}],
                "else_bodies": [],
                "expected_packages": ["foo"],
            },
        ),
        (
            "nested body followed else",
            {
                "body": [{"try": ["invalid"]}, {"else": ["bar"]}],
                "else_bodies": [],
                "expected_packages": ["bar"],
            },
        ),
        (
            "nested else followed body",
            {
                "body": ["invalid"],
                "else_bodies": [[{"try": ["foo"]}, {"else": ["bar"]}]],
                "expected_packages": ["foo"],
            },
        ),
        (
            "nested else followed else",
            {
                "body": ["invalid"],
                "else_bodies": [[{"try": ["invalid"]}, {"else": ["bar"]}]],
                "expected_packages": ["bar"],
            },
        ),
        (
            "multiple elses",
            {
                "body": ["invalid1"],
                "else_bodies": [["invalid2"], ["valid"]],
                "expected_packages": ["valid"],
            },
        ),
        (
            "multiple elses all invalid",
            {
                "body": ["invalid1"],
                "else_bodies": [["invalid2"], ["invalid3"]],
                "expected_packages": ["invalid3"],
            },
        ),
    ]

    def test_try_statement_grammar(self, body, else_bodies, expected_packages):
        processor = grammar.GrammarProcessor(
            None, snapcraft.ProjectOptions(), lambda x: "invalid" not in x
        )
        statement = _try.TryStatement(body=body, processor=processor)

        for else_body in else_bodies:
            statement.add_else(else_body)

        assert statement.process() == expected_packages


def test_else_fail():
    processor = grammar.GrammarProcessor(
        None, snapcraft.ProjectOptions(), lambda x: "invalid" not in x
    )
    statement = _try.TryStatement(body=["invalid"], processor=processor)

    statement.add_else(None)

    with pytest.raises(grammar.errors.UnsatisfiedStatementError) as error:
        statement.process()

    assert "Unable to satisfy 'try', failure forced" in str(error.value)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright 2021 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from typing import Union

from . import errors  # noqa: F401
from ._candid_client import CandidClient  # noqa: F401
from ._ubuntu_sso_client import UbuntuOneAuthClient  # noqa: F401
from ._http_client import Client  # noqa: F401


AuthClient = Union[CandidClient, UbuntuOneAuthClient]



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2018-2019 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import pathlib
from typing import Dict, Optional
from unittest import mock

from snapcraft.internal.build_providers._base_provider import Provider
from snapcraft.internal.meta.snap import Snap
from snapcraft.project import Project
from tests import fixture_setup, unit


class ProviderImpl(Provider):
    def __init__(
        self, *, project, echoer, is_ephemeral=False, build_provider_flags=None
    ):
        super().__init__(
            project=project,
            echoer=echoer,
            is_ephemeral=is_ephemeral,
            build_provider_flags=build_provider_flags,
        )

        self.run_mock = mock.Mock()
        self.launch_mock = mock.Mock()
        self.start_mock = mock.Mock()
        self.is_mounted_mock = mock.Mock(return_value=False)
        self.mount_mock = mock.Mock()
        self.unmount_mock = mock.Mock()
        self.push_file_mock = mock.Mock()
        self.create_mock = mock.Mock()
        self.destroy_mock = mock.Mock()
        self.clean_project_mock = mock.Mock()
        self.shell_mock = mock.Mock()
        self.save_info_mock = mock.Mock()
        self.loaded_info: Optional[Dict[str, str]] = None

    def _load_info(self) -> Dict[str, str]:
        if self.loaded_info is None:
            return super()._load_info()
        return self.loaded_info

    def _run(self, command, hide_output=False) -> Optional[bytes]:
        return self.run_mock(command)

    def _launch(self) -> None:
        self.launch_mock()

    def _start(self) -> None:
        self.start_mock()

    def _is_mounted(self, target: str) -> bool:
        return self.is_mounted_mock(target)

    def _mount(self, host_source: str, target: str) -> None:
        self.mount_mock(host_source, target)

    def _unmount(self, *, mountpoint: str) -> None:
        self.unmount_mock(mountpoint=mountpoint)

    def _push_file(self, *, source: str, destination: str) -> None:
        self.push_file_mock(source=source, destination=destination)

    @classmethod
    def get_instance_type_friendly_name(cls) -> str:
        return "fake-instance"

    @classmethod
    def ensure_provider(cls) -> None:
        """Fake provider check."""

    @classmethod
    def setup_provider(cls, *, echoer=None) -> None:
        """Fake provider setup."""

    @classmethod
    def _get_is_snap_injection_capable(cls) -> bool:
        return True

    @classmethod
    def _get_provider_name(cls) -> str:
        return "stub-provider"

    def _umount(self):
        raise NotImplementedError("test stub not implemented")

    def _save_info(self, **data) -> None:
        self.save_info_mock(data)

    def build_project(self) -> None:
        raise NotImplementedError("test stub not implemented")

    def create(self):
        self.create_mock("create")

    def destroy(self):
        self.destroy_mock("destroy")

    def _get_home_directory(self) -> pathlib.Path:
        return pathlib.Path("/root")

    def clean_project(self):
        self.clean_project_mock()

    def provision_project(self):
        raise NotImplementedError("test stub not implemented")

    def retrieve_snap(self):
        raise NotImplementedError("test stub not implemented")

    def pull_file(self, name: str, destination: str, delete: bool = False):
        raise NotImplementedError("test stub not implemented")

    def shell(self):
        self.shell_mock("shell")


def get_project(base: str = "core16") -> Project:
    project = Project()
    project._snap_meta = Snap(
        name="project-name", base=base, version="1.0", confinement="strict"
    )
    return project


class BaseProviderBaseTest(unit.TestCase):
    def setUp(self):
        super().setUp()

        self.useFixture(fixture_setup.FakeSnapcraftIsASnap())

        self.instance_name = "snapcraft-project-name"

        patcher = mock.patch(
            "snapcraft.internal.build_providers._base_provider.SnapInjector"
        )
        self.snap_injector_mock = patcher.start()
        self.addCleanup(patcher.stop)

        self.project = get_project()

        self.echoer_mock = mock.Mock()


class MacBaseProviderWithBasesBaseTest(unit.TestCase):
    def setUp(self):
        super().setUp()

        self.instance_name = "snapcraft-project-name"

        patcher = mock.patch(
            "snapcraft.internal.build_providers._base_provider.SnapInjector"
        )
        self.snap_injector_mock = patcher.start()
        self.addCleanup(patcher.stop)

        self.project = get_project()

        patcher = mock.patch(
            "snapcraft.internal.build_providers._base_provider._get_platform",
            return_value="darwin",
        )
        patcher.start()
        self.addCleanup(patcher.stop)

        self.echoer_mock = mock.Mock()



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2017-2021 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from textwrap import dedent

import pytest

from snapcraft.storeapi.v2 import whoami
from snapcraft.storeapi import StoreClient


@pytest.fixture
def fake_dashboard_whoami(monkeypatch):
    monkeypatch.setattr(
        StoreClient,
        "whoami",
        lambda x: whoami.WhoAmI(
            account=whoami.Account(
                email="foo@bar.baz",
                account_id="1234567890",
                name="Foo from Baz",
                username="foo",
            )
        ),
    )


@pytest.mark.usefixtures("fake_dashboard_whoami")
def test_whoami(click_run):
    result = click_run(["whoami"])

    assert result.exit_code == 0
    assert result.output == dedent(
        """\
        email:        foo@bar.baz
        developer-id: 1234567890
        """
    )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2017-2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import textwrap
from subprocess import CalledProcessError
from typing import List

from snapcraft.internal import errors, pluginhandler, steps
from snapcraft.internal.project_loader import errors as project_loader_errors
from snapcraft.internal.project_loader.inspection import errors as inspection_errors
from snapcraft.internal.repo import errors as repo_errors


def test_details_from_called_process_error():
    error = CalledProcessError(
        -1, ["test-command", "flags", "quote$me"], "test stdout", "test stderr"
    )

    details = errors.details_from_called_process_error(error)

    assert details == textwrap.dedent(
        """\
            * Command that failed: "test-command flags 'quote$me'"
            * Command exit code: -1
            * Command output: 'test stdout'
            * Command standard error output: 'test stderr'"""
    )


def test_details_from_command_error():
    details = errors.details_from_command_error(
        returncode=-1, cmd=["test-command", "flags", "quote$me"],
    )

    assert details == textwrap.dedent(
        """\
            * Command that failed: "test-command flags 'quote$me'"
            * Command exit code: -1"""
    )


def test_details_from_command_error_with_output_strings():
    details = errors.details_from_command_error(
        returncode=-1,
        cmd=["test-command", "flags", "quote$me"],
        stdout="test stdout",
        stderr="test stderr",
    )

    assert details == textwrap.dedent(
        """\
            * Command that failed: "test-command flags 'quote$me'"
            * Command exit code: -1
            * Command output: 'test stdout'
            * Command standard error output: 'test stderr'"""
    )


def test_details_from_command_error_with_output_bytes():
    details = errors.details_from_command_error(
        returncode=-1,
        cmd=["test-command", "flags", "quote$me"],
        stdout=bytes.fromhex("00 FF"),
        stderr=bytes.fromhex("01 FE"),
    )

    assert details == textwrap.dedent(
        """\
            * Command that failed: "test-command flags 'quote$me'"
            * Command exit code: -1
            * Command output: b'\\x00\\xff'
            * Command standard error output: b'\\x01\\xfe'"""
    )


class TestErrorFormatting:
    scenarios = (
        (
            "IncompatibleBaseError",
            {
                "exception_class": errors.IncompatibleBaseError,
                "kwargs": {
                    "base": "core18",
                    "linker_version": "2.23",
                    "file_list": dict(a="2.26", b="2.27"),
                },
                "expected_message": (
                    "The linker version '2.23' used by the base 'core18' is "
                    "incompatible with files in this snap:\n"
                    "    a (2.26)\n"
                    "    b (2.27)"
                ),
            },
        ),
        (
            "PrimeFileConflictError",
            {
                "exception_class": errors.PrimeFileConflictError,
                "kwargs": {"fileset": {"test-file"}},
                "expected_message": (
                    "Failed to filter files: "
                    "The following files have been excluded by the `stage` "
                    "keyword, but included by the `prime` keyword: "
                    "{'test-file'}. "
                    "Edit the `snapcraft.yaml` to make sure that the files "
                    "included in `prime` are also included in `stage`."
                ),
            },
        ),
        (
            "SnapcraftAfterPartMissingError",
            {
                "exception_class": project_loader_errors.SnapcraftAfterPartMissingError,
                "kwargs": {"part_name": "test-part1", "after_part_name": "test-part2"},
                "expected_message": (
                    "Failed to get part information: "
                    "Cannot find the definition for part 'test-part2', required by "
                    "part 'test-part1'.\n"
                    "Remote parts are not supported with bases, so make sure that this "
                    "part is defined in the `snapcraft.yaml`."
                ),
            },
        ),
        (
            "PluginError",
            {
                "exception_class": errors.PluginError,
                "kwargs": {"message": "test-message"},
                "expected_message": "Failed to load plugin: test-message",
            },
        ),
        (
            "PluginBaseError",
            {
                "exception_class": errors.PluginBaseError,
                "kwargs": {"part_name": "go-part", "base": "arch"},
                "expected_message": "The plugin used by part 'go-part' does not support snaps using base 'arch'.",
            },
        ),
        (
            "SnapcraftPartConflictError",
            {
                "exception_class": errors.SnapcraftPartConflictError,
                "kwargs": {
                    "part_name": "test-part",
                    "other_part_name": "test-other-part",
                    "conflict_files": ("test-file1", "test-file2"),
                },
                "expected_message": (
                    "Failed to stage: "
                    "Parts 'test-other-part' and 'test-part' have the following "
                    "files, but with different contents:\n"
                    "    test-file1\n"
                    "    test-file2\n"
                    "\n"
                    "Snapcraft offers some capabilities to solve this by use of "
                    "the following keywords:\n"
                    "    - `filesets`\n"
                    "    - `stage`\n"
                    "    - `snap`\n"
                    "    - `organize`\n"
                    "\n"
                    "To learn more about these part keywords, run "
                    "`snapcraft help plugins`."
                ),
            },
        ),
        (
            "InvalidWikiEntryError",
            {
                "exception_class": errors.InvalidWikiEntryError,
                "kwargs": {"error": "test-error"},
                "expected_message": "Invalid wiki entry: 'test-error'",
            },
        ),
        (
            "PluginOutdatedError",
            {
                "exception_class": errors.PluginOutdatedError,
                "kwargs": {"message": "test-message"},
                "expected_message": "This plugin is outdated: test-message",
            },
        ),
        (
            "RequiredCommandFailure",
            {
                "exception_class": errors.RequiredCommandFailure,
                "kwargs": {"command": "test-command"},
                "expected_message": "'test-command' failed.",
            },
        ),
        (
            "RequiredCommandNotFound",
            {
                "exception_class": errors.RequiredCommandNotFound,
                "kwargs": {"cmd_list": ["test-command", "test-argument"]},
                "expected_message": "'test-command' not found.",
            },
        ),
        (
            "RequiredPathDoesNotExist",
            {
                "exception_class": errors.RequiredPathDoesNotExist,
                "kwargs": {"path": "test-path"},
                "expected_message": "Required path does not exist: 'test-path'",
            },
        ),
        (
            "SnapcraftPathEntryError",
            {
                "exception_class": errors.SnapcraftPathEntryError,
                "kwargs": {"value": "test-path", "key": "test-key", "app": "test-app"},
                "expected_message": (
                    "Failed to generate snap metadata: "
                    "The path 'test-path' set for 'test-key' in 'test-app' does "
                    "not exist. Make sure that the files are in the `prime` "
                    "directory."
                ),
            },
        ),
        (
            "InvalidPullPropertiesError",
            {
                "exception_class": errors.InvalidPullPropertiesError,
                "kwargs": {
                    "plugin_name": "test-plugin",
                    "properties": ["test-property1", "test-property2"],
                },
                "expected_message": (
                    "Failed to load plugin: "
                    "Invalid pull properties specified by 'test-plugin' plugin: "
                    "['test-property1', 'test-property2']"
                ),
            },
        ),
        (
            "InvalidBuildPropertiesError",
            {
                "exception_class": errors.InvalidBuildPropertiesError,
                "kwargs": {
                    "plugin_name": "test-plugin",
                    "properties": ["test-property1", "test-property2"],
                },
                "expected_message": (
                    "Failed to load plugin: "
                    "Invalid build properties specified by 'test-plugin' plugin: "
                    "['test-property1', 'test-property2']"
                ),
            },
        ),
        (
            "StagePackageDownloadError",
            {
                "exception_class": errors.StagePackageDownloadError,
                "kwargs": {"part_name": "test-part", "message": "test-message"},
                "expected_message": (
                    "Failed to fetch stage packages: "
                    "Error downloading packages for part "
                    "'test-part': test-message."
                ),
            },
        ),
        (
            "InvalidContainerImageInfoError",
            {
                "exception_class": errors.InvalidContainerImageInfoError,
                "kwargs": {"image_info": "test-image-info"},
                "expected_message": (
                    "Failed to parse container image info: "
                    "SNAPCRAFT_IMAGE_INFO is not a valid JSON string: "
                    "test-image-info"
                ),
            },
        ),
        (
            "MissingMetadataFileError",
            {
                "exception_class": errors.MissingMetadataFileError,
                "kwargs": {"part_name": "test-part", "path": "test/path"},
                "expected_message": (
                    "Failed to generate snap metadata: "
                    "Part 'test-part' has a 'parse-info' referring to metadata "
                    "file 'test/path', which does not exist."
                ),
            },
        ),
        (
            "UnhandledMetadataFileTypeError",
            {
                "exception_class": errors.UnhandledMetadataFileTypeError,
                "kwargs": {"path": "test/path"},
                "expected_message": (
                    "Failed to extract metadata from 'test/path': "
                    "This type of file is not supported for supplying "
                    "metadata."
                ),
            },
        ),
        (
            "InvalidExtractorValueError",
            {
                "exception_class": errors.InvalidExtractorValueError,
                "kwargs": {"path": "test/path", "extractor_name": "extractor"},
                "expected_message": (
                    "Failed to extract metadata from 'test/path': "
                    "Extractor 'extractor' didn't return ExtractedMetadata as "
                    "expected."
                ),
            },
        ),
        (
            "PatcherNewerPatchelfError",
            {
                "exception_class": errors.PatcherNewerPatchelfError,
                "kwargs": {
                    "elf_file": "test/path",
                    "patchelf_version": "patchelf 0.9",
                    "process_exception": CalledProcessError(
                        cmd=["patchelf"], returncode=-1
                    ),
                },
                "expected_message": (
                    "'test/path' cannot be patched to function properly in a "
                    "classic confined snap: patchelf failed with exit code -1.\n"
                    "'patchelf 0.9' may be too old. A newer version of patchelf "
                    "may be required.\n"
                    "Try adding the `after: [patchelf]` and a `patchelf` part "
                    "that would filter out files from prime `prime: [-*]` or "
                    "`build-snaps: [patchelf/latest/edge]` to the failing part "
                    "in your `snapcraft.yaml` to use a newer patchelf."
                ),
            },
        ),
        (
            "PatcherGenericError",
            {
                "exception_class": errors.PatcherGenericError,
                "kwargs": {
                    "elf_file": "test/path",
                    "process_exception": CalledProcessError(
                        cmd=["patchelf"], returncode=-1
                    ),
                },
                "expected_message": (
                    "'test/path' cannot be patched to function properly in a "
                    "classic confined snap: patchelf failed with exit code -1"
                ),
            },
        ),
        (
            "StagePackageMissingError",
            {
                "exception_class": errors.StagePackageMissingError,
                "kwargs": {"package": "libc6"},
                "expected_message": (
                    "'libc6' is required inside the snap for this "
                    "part to work properly.\nAdd it as a `stage-packages` "
                    "entry for this part."
                ),
            },
        ),
        (
            "SnapcraftCommandError",
            {
                "exception_class": errors.SnapcraftCommandError,
                "kwargs": {
                    "command": "pip install foo",
                    "call_error": CalledProcessError(
                        cmd=["/bin/sh"], returncode=1, output="failed"
                    ),
                },
                "expected_message": (
                    "Failed to run 'pip install foo': Exited with code 1."
                ),
            },
        ),
        (
            "SnapcraftPluginCommandError string command",
            {
                "exception_class": errors.SnapcraftPluginCommandError,
                "kwargs": {
                    "command": "make install",
                    "exit_code": -1,
                    "part_name": "make_test",
                },
                "expected_message": (
                    "Failed to run 'make install' for 'make_test': "
                    "Exited with code -1.\n"
                    "Verify that the part is using the correct parameters and try "
                    "again."
                ),
            },
        ),
        (
            "SnapcraftPluginCommandError list command",
            {
                "exception_class": errors.SnapcraftPluginCommandError,
                "kwargs": {
                    "command": ["make", "install"],
                    "exit_code": 2,
                    "part_name": "make_test",
                },
                "expected_message": (
                    "Failed to run 'make install' for 'make_test': "
                    "Exited with code 2.\n"
                    "Verify that the part is using the correct parameters and try "
                    "again."
                ),
            },
        ),
        (
            "CrossCompilationNotSupported",
            {
                "exception_class": errors.CrossCompilationNotSupported,
                "kwargs": {"part_name": "my-part"},
                "expected_message": (
                    "The plugin used by 'my-part' does not support "
                    "cross-compiling to a different target architecture."
                ),
            },
        ),
        (
            "CacheUpdateFailedError",
            {
                "exception_class": repo_errors.CacheUpdateFailedError,
                "kwargs": {"errors": ""},
                "expected_message": (
                    "Failed to update the package cache: "
                    "Some files could not be downloaded: "
                    "Check that the sources on your host are configured correctly."
                ),
            },
        ),
        (
            "CacheUpdateFailedError",
            {
                "exception_class": repo_errors.CacheUpdateFailedError,
                "kwargs": {"errors": "foo, bar"},
                "expected_message": (
                    "Failed to update the package cache: "
                    "Some files could not be downloaded:\n\nfoo\nbar\n\n"
                    "Check that the sources on your host are configured correctly."
                ),
            },
        ),
        (
            "SnapcraftCopyFileNotFoundError",
            {
                "exception_class": errors.SnapcraftCopyFileNotFoundError,
                "kwargs": {"path": "test-path"},
                "expected_message": (
                    "Failed to copy 'test-path': no such file or directory.\n"
                    "Check the path and try again."
                ),
            },
        ),
        (
            "MountPointNotFoundError",
            {
                "exception_class": errors.MountPointNotFoundError,
                "kwargs": {"mount_point": "test-mount-point"},
                "expected_message": "Nothing is mounted at 'test-mount-point'",
            },
        ),
        (
            "RootNotMountedError",
            {
                "exception_class": errors.RootNotMountedError,
                "kwargs": {"root": "test-root"},
                "expected_message": "'test-root' is not mounted",
            },
        ),
        (
            "InvalidMountinfoFormat",
            {
                "exception_class": errors.InvalidMountinfoFormat,
                "kwargs": {"row": [1, 2, 3]},
                "expected_message": "Unable to parse mountinfo row: [1, 2, 3]",
            },
        ),
        (
            "InvalidStepError",
            {
                "exception_class": errors.InvalidStepError,
                "kwargs": {"step_name": "test-step-name"},
                "expected_message": "'test-step-name' is not a valid lifecycle step",
            },
        ),
        (
            "NoLatestStepError",
            {
                "exception_class": errors.NoLatestStepError,
                "kwargs": {"part_name": "test-part-name"},
                "expected_message": "The 'test-part-name' part hasn't run any steps",
            },
        ),
        (
            "NoNextStepError",
            {
                "exception_class": errors.NoNextStepError,
                "kwargs": {"part_name": "test-part-name"},
                "expected_message": (
                    "The 'test-part-name' part has run through its entire lifecycle"
                ),
            },
        ),
        (
            "StepHasNotRunError",
            {
                "exception_class": errors.StepHasNotRunError,
                "kwargs": {"part_name": "test-part-name", "step": steps.BUILD},
                "expected_message": (
                    "The 'test-part-name' part has not yet run the 'build' step"
                ),
            },
        ),
        (
            "ScriptletDuplicateFieldError",
            {
                "exception_class": errors.ScriptletDuplicateFieldError,
                "kwargs": {"field": "foo", "step": steps.PULL},
                "expected_message": (
                    "Unable to set foo: it was already set in the 'pull' step."
                ),
            },
        ),
        (
            "ToolMissingError",
            {
                "exception_class": errors.ToolMissingError,
                "kwargs": {"command_name": "runnable"},
                "expected_message": (
                    "A tool snapcraft depends on could not be found: 'runnable'.\n"
                    "Ensure the tool is installed and available, and try again."
                ),
            },
        ),
        (
            "NoSuchFileError",
            {
                "exception_class": inspection_errors.NoSuchFileError,
                "kwargs": {"path": "test-path"},
                "expected_message": (
                    "Failed to find part that provided path: 'test-path' does not "
                    "exist.\n"
                    "Check the file path and try again."
                ),
            },
        ),
        (
            "ProvidesInvalidFilePathError",
            {
                "exception_class": inspection_errors.ProvidesInvalidFilePathError,
                "kwargs": {"path": "test-path"},
                "expected_message": (
                    "Failed to find part that provides path: 'test-path' is not "
                    "in the staging or priming area.\n"
                    "Ensure the path is in the staging or priming area and try "
                    "again."
                ),
            },
        ),
        (
            "UntrackedFileError",
            {
                "exception_class": inspection_errors.UntrackedFileError,
                "kwargs": {"path": "test-path"},
                "expected_message": (
                    "No known parts provided 'test-path'. It may have been "
                    "provided by a scriptlet."
                ),
            },
        ),
        (
            "NoStepsRunError",
            {
                "exception_class": inspection_errors.NoStepsRunError,
                "kwargs": {},
                "expected_message": "Failed to get latest step: no steps have run",
            },
        ),
    )

    def test_error_formatting(self, exception_class, expected_message, kwargs):
        assert str(exception_class(**kwargs)) == expected_message


class StrangeExceptionSimple(errors.SnapcraftException):
    def get_brief(self):
        return "something's strange, in the neighborhood"

    def get_resolution(self):
        return "who you gonna call? ghostbusters!!"

    def get_details(self):
        return "i ain't afraid of no ghosts"

    def get_docs_url(self):
        return "https://docs.snapcraft.io/the-snapcraft-format/8337"


class StrangeExceptionWithFormatting(errors.SnapcraftException):
    def __init__(self, neighborhood: str, contact: str, ghosts: List[str]) -> None:
        self._neighborhood = neighborhood
        self._contact = contact
        self._ghosts = ghosts

    def get_brief(self):
        return "something's strange, in the neighborhood of {}".format(
            self._neighborhood
        )

    def get_resolution(self):
        return "who you gonna call? {}!!".format(self._contact)

    def get_details(self):
        return "i ain't afraid of no ghosts: {}".format(self._ghosts)

    def get_docs_url(self):
        return "https://docs.snapcraft.io/the-snapcraft-format/8337"


class TestSnapcraftExceptionTests:

    scenarios = (
        (
            "StrangeExceptionSimple",
            {
                "exception_class": StrangeExceptionSimple,
                "kwargs": {},
                "expected_brief": "something's strange, in the neighborhood",
                "expected_resolution": "who you gonna call? ghostbusters!!",
                "expected_details": "i ain't afraid of no ghosts",
                "expected_docs_url": "https://docs.snapcraft.io/the-snapcraft-format/8337",
                "expected_reportable": False,
            },
        ),
        (
            "StrangeExceptionWithFormatting",
            {
                "exception_class": StrangeExceptionWithFormatting,
                "kwargs": {
                    "neighborhood": "Times Square",
                    "ghosts": ["slimer", "puft", "vigo"],
                    "contact": "Janine Melnitz",
                },
                "expected_brief": "something's strange, in the neighborhood of Times Square",
                "expected_resolution": "who you gonna call? Janine Melnitz!!",
                "expected_details": "i ain't afraid of no ghosts: ['slimer', 'puft', 'vigo']",
                "expected_docs_url": "https://docs.snapcraft.io/the-snapcraft-format/8337",
                "expected_reportable": False,
            },
        ),
        (
            "SnapcraftHostToolNotFoundError",
            {
                "exception_class": errors.SnapcraftHostToolNotFoundError,
                "kwargs": {"command_name": "foo", "package_name": "foo-pkg"},
                "expected_brief": "A tool snapcraft depends on could not be found: 'foo'",
                "expected_resolution": "Ensure that 'foo-pkg' is installed.",
                "expected_details": None,
                "expected_docs_url": None,
                "expected_reportable": False,
            },
        ),
        (
            "MissingStateCleanError",
            {
                "exception_class": errors.MissingStateCleanError,
                "kwargs": {"step": steps.PULL},
                "expected_brief": "Failed to clean for step 'pull'.",
                "expected_resolution": "Run `snapcraft clean` and retry build.",
                "expected_details": None,
                "expected_docs_url": None,
                "expected_reportable": False,
            },
        ),
        (
            "StepOutdatedError dirty_properties",
            {
                "exception_class": errors.StepOutdatedError,
                "kwargs": {
                    "step": steps.PULL,
                    "part": "test-part",
                    "dirty_report": pluginhandler.DirtyReport(
                        dirty_properties=["test-property1", "test-property2"]
                    ),
                },
                "expected_brief": "Failed to reuse files from previous run.",
                "expected_resolution": "Run `snapcraft clean` and retry build.",
                "expected_details": "The 'test-property1' and 'test-property2' part properties appear to have changed.\n",
                "expected_docs_url": None,
                "expected_reportable": False,
            },
        ),
        (
            "StepOutdatedError dirty_project_options",
            {
                "exception_class": errors.StepOutdatedError,
                "kwargs": {
                    "step": steps.PULL,
                    "part": "test-part",
                    "dirty_report": pluginhandler.DirtyReport(
                        dirty_project_options=["test-option"]
                    ),
                },
                "expected_brief": "Failed to reuse files from previous run.",
                "expected_resolution": "Run `snapcraft clean` and retry build.",
                "expected_details": "The 'test-option' project option appears to have changed.\n",
                "expected_docs_url": None,
                "expected_reportable": False,
            },
        ),
        (
            "StepOutdatedError changed_dependencies",
            {
                "exception_class": errors.StepOutdatedError,
                "kwargs": {
                    "step": steps.PULL,
                    "part": "test-part",
                    "dirty_report": pluginhandler.DirtyReport(
                        changed_dependencies=[
                            pluginhandler.Dependency(
                                part_name="another-part", step=steps.PULL
                            )
                        ]
                    ),
                },
                "expected_brief": "Failed to reuse files from previous run.",
                "expected_resolution": "Run `snapcraft clean` and retry build.",
                "expected_details": "A dependency has changed: 'another-part'\n",
                "expected_docs_url": None,
                "expected_reportable": False,
            },
        ),
        (
            "StepOutdatedError multiple changed_dependencies",
            {
                "exception_class": errors.StepOutdatedError,
                "kwargs": {
                    "step": steps.PULL,
                    "part": "test-part",
                    "dirty_report": pluginhandler.DirtyReport(
                        changed_dependencies=[
                            pluginhandler.Dependency(
                                part_name="another-part1", step=steps.PULL
                            ),
                            pluginhandler.Dependency(
                                part_name="another-part2", step=steps.PULL
                            ),
                        ]
                    ),
                },
                "expected_brief": "Failed to reuse files from previous run.",
                "expected_resolution": "Run `snapcraft clean` and retry build.",
                "expected_details": "Some dependencies have changed: 'another-part1' and 'another-part2'\n",
                "expected_docs_url": None,
                "expected_reportable": False,
            },
        ),
        (
            "StepOutdatedError previous step updated",
            {
                "exception_class": errors.StepOutdatedError,
                "kwargs": {
                    "step": steps.STAGE,
                    "part": "test-part",
                    "outdated_report": pluginhandler.OutdatedReport(
                        previous_step_modified=steps.BUILD
                    ),
                },
                "expected_brief": "Failed to reuse files from previous run.",
                "expected_resolution": "Run `snapcraft clean` and retry build.",
                "expected_details": "The 'build' step has run more recently.\n",
                "expected_docs_url": None,
                "expected_reportable": False,
            },
        ),
        (
            "StepOutdatedError source updated",
            {
                "exception_class": errors.StepOutdatedError,
                "kwargs": {
                    "step": steps.PULL,
                    "part": "test-part",
                    "outdated_report": pluginhandler.OutdatedReport(
                        source_updated=True
                    ),
                },
                "expected_brief": "Failed to reuse files from previous run.",
                "expected_resolution": "Run `snapcraft clean` and retry build.",
                "expected_details": "The source has changed on disk.\n",
                "expected_docs_url": None,
                "expected_reportable": False,
            },
        ),
        (
            "SnapcraftEnvironmentError",
            {
                "exception_class": errors.SnapcraftEnvironmentError,
                "kwargs": {"message": "test-message"},
                "expected_brief": "test-message",
                "expected_resolution": "",
                "expected_details": None,
                "expected_docs_url": None,
                "expected_reportable": False,
            },
        ),
        (
            "SnapcraftDataDirectoryMissingError",
            {
                "exception_class": errors.SnapcraftDataDirectoryMissingError,
                "kwargs": {},
                "expected_brief": "Cannot find snapcraft's data files.",
                "expected_resolution": "Re-install snapcraft or verify installation is correct.",
                "expected_details": None,
                "expected_docs_url": None,
                "expected_reportable": True,
            },
        ),
        (
            "SnapcraftMissingLinkerInBaseError",
            {
                "exception_class": errors.SnapcraftMissingLinkerInBaseError,
                "kwargs": {
                    "base": "core18",
                    "linker_path": "/snap/core18/current/lib64/ld-linux.so.2",
                },
                "expected_brief": "Cannot find the linker to use for the target base 'core18'.",
                "expected_resolution": "Verify that the linker exists at the expected path '/snap/core18/current/lib64/ld-linux.so.2' and try again. If the linker does not exist contact the author of the base (run `snap info core18` to get information for this base).",
                "expected_details": None,
                "expected_docs_url": None,
                "expected_reportable": False,
            },
        ),
        (
            "SnapcraftPluginAssertionError",
            {
                "exception_class": errors.SnapcraftPluginAssertionError,
                "kwargs": {"name": "part-name", "reason": "missing important file"},
                "expected_brief": "Unable to build 'part-name': missing important file",
                "expected_resolution": "Ensure the part's configuration and sources are correct.",
                "expected_details": "",
                "expected_docs_url": None,
                "expected_reportable": False,
            },
        ),
        (
            "XAttributeError",
            {
                "exception_class": errors.XAttributeError,
                "kwargs": {"path": "/tmp/foo", "key": "foo", "action": "read"},
                "expected_brief": "Unable to read extended attribute.",
                "expected_resolution": "Check that your filesystem supports extended attributes.",
                "expected_details": "Failed to read attribute 'foo' on '/tmp/foo'.",
                "expected_docs_url": None,
                "expected_reportable": True,
            },
        ),
        (
            "XAttributeTooLongError",
            {
                "exception_class": errors.XAttributeTooLongError,
                "kwargs": {"path": "/tmp/foo", "key": "foo", "value": "bar"},
                "expected_brief": "Unable to write extended attribute as the key and/or value is too long.",
                "expected_resolution": "This issue is generally resolved by addressing/truncating the data source of the long data value. In some cases, the filesystem being used will limit the allowable size.",
                "expected_details": "Failed to write attribute to '/tmp/foo':\nkey='foo' value='bar'",
                "expected_docs_url": None,
                "expected_reportable": True,
            },
        ),
        (
            "SnapcraftPluginBuildError",
            {
                "exception_class": errors.SnapcraftPluginBuildError,
                "kwargs": {"part_name": "foo"},
                "expected_brief": "Failed to build 'foo'.",
                "expected_resolution": "Check the build logs and ensure the part's configuration and sources are correct.",
                "expected_details": None,
                "expected_docs_url": None,
                "expected_reportable": False,
            },
        ),
    )

    def test_snapcraft_exception_handling(
        self,
        exception_class,
        expected_brief,
        expected_details,
        expected_docs_url,
        expected_reportable,
        expected_resolution,
        kwargs,
    ):
        exception = exception_class(**kwargs)

        assert exception.get_brief() == expected_brief
        assert exception.get_resolution() == expected_resolution
        assert exception.get_details() == expected_details
        assert exception.get_docs_url() == expected_docs_url
        assert exception.get_reportable() == expected_reportable



import setuptools

# if dependencies are not handled first, processing of this
# setup.py will fail.
import six  # noqa: F401

setuptools.setup(
    name="hello-world",
    version="0.0.1",
    author="Canonical LTD",
    author_email="snapcraft@lists.snapcraft.io",
    description="A simple hello world in python",
    scripts=["hello"],
)



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2017-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os

from testtools.matchers import Contains, FileContains, FileExists, Not

from tests import integration


class HookTestCase(integration.TestCase):
    def test_hooks(self):
        self.run_snapcraft("prime", "hooks")

        for hook in ("configure", "another-hook"):
            # Assert that the hooks as supplied to snapcraft was copied into
            # the snap.
            self.assertThat(
                os.path.join(self.prime_dir, "snap", "hooks", hook), FileExists()
            )

            # Assert that the real hooks were generated as well (they're just
            # wrappers that call the one given to snapcraft).
            self.assertThat(
                os.path.join(self.prime_dir, "meta", "hooks", hook), FileExists()
            )

        # Assert that the wrapper execs the correct thing
        self.assertThat(
            os.path.join(self.prime_dir, "meta", "hooks", "another-hook"),
            FileContains(matcher=Contains('exec "$SNAP/snap/hooks/{}"'.format(hook))),
        )

        # Assert that the configure hook doesn't have a wrapper
        self.assertThat(
            os.path.join(self.prime_dir, "meta", "hooks", "configure"),
            Not(FileContains(matcher=Contains("exec"))),
        )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2016 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import functools
import logging
import os

from progressbar import Bar, Percentage, ProgressBar
from requests_toolbelt import MultipartEncoder, MultipartEncoderMonitor

from snapcraft.storeapi.errors import StoreUpDownError

logger = logging.getLogger(__name__)


def _update_progress_bar(progress_bar, maximum_value, monitor):
    if monitor.bytes_read <= maximum_value:
        progress_bar.update(monitor.bytes_read)


def upload_files(binary_filename, updown_client):
    """Upload a binary file to the Store.

    Submit a file to the Store upload service and return the
    corresponding upload_id.
    """
    try:
        binary_file_size = os.path.getsize(binary_filename)
        binary_file = open(binary_filename, "rb")
        encoder = MultipartEncoder(
            fields={"binary": ("filename", binary_file, "application/octet-stream")}
        )

        # Create a progress bar that looks like: Uploading foo [==  ] 50%
        progress_bar = ProgressBar(
            widgets=[
                "Pushing {!r} ".format(os.path.basename(binary_filename)),
                Bar(marker="=", left="[", right="]"),
                " ",
                Percentage(),
            ],
            maxval=os.path.getsize(binary_filename),
        )
        progress_bar.start()
        # Create a monitor for this upload, so that progress can be displayed
        monitor = MultipartEncoderMonitor(
            encoder,
            functools.partial(_update_progress_bar, progress_bar, binary_file_size),
        )

        # Begin upload
        response = updown_client.upload(monitor)

        # Make sure progress bar shows 100% complete
        progress_bar.finish()
    finally:
        # Close the open file
        binary_file.close()

    if not response.ok:
        raise StoreUpDownError(response)

    response_data = response.json()
    return {
        "upload_id": response_data["upload_id"],
        "binary_filesize": binary_file_size,
        "source_uploaded": False,
    }



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import contextlib
from urllib import parse

import apt
import requests_unixsocket
from requests import exceptions


def is_snap_installed(snap):
    return get_local_snap_info(snap) is not None


def get_local_snap_info(snap):
    local_snap_info = None
    with contextlib.suppress(exceptions.HTTPError):
        local_snap_info = _get_local_snap_info(_get_parsed_snap(snap)[0])
    return local_snap_info


def _get_parsed_snap(snap):
    if "/" in snap:
        sep_index = snap.find("/")
        snap_name = snap[:sep_index]
        snap_channel = snap[sep_index + 1 :]
    else:
        snap_name = snap
        snap_channel = ""
    return snap_name, snap_channel


def _get_snapd_socket_path_template():
    return "http+unix://%2Frun%2Fsnapd.socket/v2/{}"


def _get_local_snap_info(name):
    slug = "snaps/{}".format(parse.quote(name, safe=""))
    url = _get_snapd_socket_path_template().format(slug)
    with requests_unixsocket.Session() as session:
        snap_info = session.get(url)
    snap_info.raise_for_status()
    return snap_info.json()["result"]


def is_package_installed(name):
    with apt.Cache() as apt_cache:
        return apt_cache[name].installed



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2016-2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os

from testtools.matchers import Contains, DirExists, FileExists, Not

from tests import integration


class CleanPrimeStepTestCase(integration.TestCase):
    def setUp(self):
        super().setUp()

        self.copy_project_to_cwd("independent-parts")
        self.run_snapcraft("prime")

    def test_clean_prime_step(self):
        bindir = os.path.join(self.prime_dir, "bin")
        self.assertThat(os.path.join(bindir, "file1"), FileExists())
        self.assertThat(os.path.join(bindir, "file2"), FileExists())

        output = self.run_snapcraft(["clean", "--step=prime"], debug=False)
        self.assertThat(self.prime_dir, Not(DirExists()))
        self.assertThat(self.stage_dir, DirExists())
        self.assertThat(self.parts_dir, DirExists())

        # Assert that the priming area was removed wholesale, not a part at a
        # time (since we didn't specify any parts).
        self.assertThat(output, Contains("Cleaning up priming area"))
        self.expectThat(output, Not(Contains("part1")))
        self.expectThat(output, Not(Contains("part2")))

        # Now try to prime again
        self.run_snapcraft("prime")
        self.assertThat(os.path.join(bindir, "file1"), FileExists())
        self.assertThat(os.path.join(bindir, "file2"), FileExists())

    def test_clean_prime_step_single_part(self):
        bindir = os.path.join(self.prime_dir, "bin")
        self.assertThat(os.path.join(bindir, "file1"), FileExists())
        self.assertThat(os.path.join(bindir, "file2"), FileExists())

        self.run_snapcraft(["clean", "part1", "--step=prime"])
        self.assertThat(os.path.join(bindir, "file1"), Not(FileExists()))
        self.assertThat(os.path.join(bindir, "file2"), FileExists())
        self.assertThat(self.stage_dir, DirExists())
        self.assertThat(self.parts_dir, DirExists())

        # Now try to prime again
        self.run_snapcraft("prime")
        self.assertThat(os.path.join(bindir, "file1"), FileExists())
        self.assertThat(os.path.join(bindir, "file2"), FileExists())

    def test_clean_with_deprecated_strip_step(self):
        bindir = os.path.join(self.prime_dir, "bin")
        self.assertThat(os.path.join(bindir, "file1"), FileExists())
        self.assertThat(os.path.join(bindir, "file2"), FileExists())

        self.run_snapcraft(["clean", "--step=strip"])
        self.assertThat(self.prime_dir, Not(DirExists()))
        self.assertThat(self.stage_dir, DirExists())
        self.assertThat(self.parts_dir, DirExists())

        # Now try to prime again
        self.run_snapcraft("prime")
        self.assertThat(os.path.join(bindir, "file1"), FileExists())
        self.assertThat(os.path.join(bindir, "file2"), FileExists())



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2019-2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import subprocess
import sys
from typing import Any, Dict
from unittest import mock

from testtools.matchers import Equals, FileContains, FileExists

from snapcraft.internal.build_providers import _base_provider, errors
from snapcraft.internal.build_providers._lxd import LXD
from snapcraft.internal.errors import SnapcraftEnvironmentError
from snapcraft.internal.repo.errors import SnapdConnectionError
from tests.unit.build_providers import BaseProviderBaseTest

if sys.platform == "linux":
    import pylxd

LXD_RUN_ENV_COMMAND = [
    "env",
    "SNAPCRAFT_HAS_TTY=False",
]

LXD_RUN_COMMAND_PREFIX = [
    "/snap/bin/lxc",
    "exec",
    "snapcraft-project-name",
    "--",
] + LXD_RUN_ENV_COMMAND


class GetEnv(_base_provider.Provider):
    def _get_env_command(self):
        return LXD_RUN_ENV_COMMAND.copy()


class EnvSetup(_base_provider.Provider):
    def _setup_environment(self):
        pass


class LXDTestImpl(LXD, GetEnv, EnvSetup):
    pass


class FakeContainer:
    @property
    def status(self) -> str:
        return self._status

    def __init__(self, config: Dict[str, Any], wait: bool) -> None:
        self._status = "STOPPED"
        self.config = config
        self.devices = dict()  # type: Dict[str, Any]

        self.delete_mock = mock.Mock()
        self.files_delete_mock = mock.Mock()
        self.files_get_mock = mock.Mock()
        self.files_put_mock = mock.Mock()
        self.restart_mock = mock.Mock()
        self.save_mock = mock.Mock()
        self.start_mock = mock.Mock()
        self.stop_mock = mock.Mock()
        self.sync_mock = mock.Mock()

        class FakeContainerFiles:
            delete_available = True

            @staticmethod
            def get(file_name) -> bytes:
                self.files_get_mock(file_name=file_name)
                return b"fake-pull"

            @staticmethod
            def put(destination: str, contents: bytes) -> None:
                self.files_put_mock(destination=destination, contents=contents)

            @staticmethod
            def delete(file_name) -> None:
                self.files_delete_mock(file_name=file_name)

        self.files = FakeContainerFiles

    def _reset_mocks(self) -> None:
        self.delete_mock.reset_mock()
        self.files_delete_mock.reset_mock()
        self.files_get_mock.reset_mock()
        self.files_put_mock.reset_mock()
        self.restart_mock.reset_mock()
        self.save_mock.reset_mock()
        self.start_mock.reset_mock()
        self.stop_mock.reset_mock()
        self.sync_mock.reset_mock()

    def restart(self, wait: bool) -> None:
        self.restart_mock(wait=wait)

    def save(self, wait: bool) -> None:
        self.save_mock(wait=wait)

    def state(self):
        class State:
            network = dict(
                eth0=dict(addresses=[dict(address="192.168.0.1", family="inet")])
            )

        return State()

    def sync(self) -> None:
        self.sync_mock()

    def start(self, wait: bool) -> None:
        self.start_mock(wait=wait)
        self._status = "RUNNING"

    def stop(self, wait: bool) -> None:
        self.stop_mock(wait=wait)
        self._status = "STOPPED"

    def delete(self, wait: bool) -> None:
        self.delete_mock(wait=wait)


class FakeContainers:
    def __init__(self):
        self._containers = dict()  # type: Dict[str, FakeContainer]
        self.create_mock = mock.Mock()
        self.get_mock = mock.Mock()

    def exists(self, container_name: str) -> bool:
        return container_name in self._containers

    def get(self, container_name: str) -> FakeContainer:
        self.get_mock(container_name)
        return self._containers[container_name]

    def create(self, config: Dict[str, Any], wait: bool) -> FakeContainer:
        self.create_mock(config=config, wait=wait)
        self._containers[config["name"]] = FakeContainer(config, wait)
        return self._containers[config["name"]]


class FakePyLXDClient:
    def __init__(self) -> None:
        self.containers = FakeContainers()
        self.host_info = {
            "environment": {"kernel_features": {"seccomp_listener": "true"}}
        }


class LXDBaseTest(BaseProviderBaseTest):
    def setUp(self):
        super().setUp()

        self.fake_pylxd_client = FakePyLXDClient()
        patcher = mock.patch("pylxd.Client", return_value=self.fake_pylxd_client)
        self.pylxd_mock = patcher.start()
        self.addCleanup(patcher.stop)

        patcher = mock.patch(
            "snapcraft.internal.build_providers._base_provider.Provider.clean_project",
            return_value=True,
        )

        patcher.start()
        self.addCleanup(patcher.stop)

        patcher = mock.patch("subprocess.check_call", spec=subprocess.check_call)
        self.check_call_mock = patcher.start()
        self.addCleanup(patcher.stop)

        def check_output_effect(command: str, *arch, **kwargs) -> bytes:
            if "is-system-running" in command:
                raise subprocess.CalledProcessError(
                    returncode=1,
                    cmd=["systemctl", "is-system-running"],
                    output="degraded\n".encode(),
                )

            if "getent" in command:
                output = "2001:67c:1562::20 snapcraft.io"
            else:
                output = ""

            return output.encode()

        patcher = mock.patch(
            "subprocess.check_output",
            spec=subprocess.check_output,
            side_effect=check_output_effect,
        )
        self.check_output_mock = patcher.start()
        self.addCleanup(patcher.stop)


class LXDInitTest(LXDBaseTest):
    def test_create(self):
        instance = LXDTestImpl(project=self.project, echoer=self.echoer_mock)

        instance.create()

        self.fake_pylxd_client.containers.create_mock.assert_called_once_with(
            config={
                "name": "snapcraft-project-name",
                "raw.idmap": f"both {os.getuid()} 0",
                "security.syscalls.intercept.mknod": "true",
                "source": {
                    "mode": "pull",
                    "type": "image",
                    "server": "https://cloud-images.ubuntu.com/buildd/releases",
                    "protocol": "simplestreams",
                    "alias": "16.04",
                },
            },
            wait=True,
        )

        container = self.fake_pylxd_client.containers.get(self.instance_name)
        container.start_mock.assert_called_once_with(wait=True)
        self.assertThat(container.save_mock.call_count, Equals(3))

        for args, kwargs in (
            self.check_output_mock.call_args_list + self.check_call_mock.call_args_list
        ):
            assert args[0][0 : len(LXD_RUN_COMMAND_PREFIX)] == LXD_RUN_COMMAND_PREFIX

        check_output_calls = [
            args[0][len(LXD_RUN_COMMAND_PREFIX) :]
            for args, kwargs in self.check_output_mock.call_args_list
        ]

        assert check_output_calls == [
            ["hostname", "-F", "/etc/hostname"],
            ["systemctl", "is-system-running"],
            ["ln", "-sf", "/run/systemd/resolve/resolv.conf", "/etc/resolv.conf"],
            ["systemctl", "enable", "systemd-resolved"],
            ["systemctl", "enable", "systemd-networkd"],
            ["systemctl", "restart", "systemd-resolved"],
            ["systemctl", "restart", "systemd-networkd"],
            ["getent", "hosts", "snapcraft.io"],
            ["systemctl", "enable", "systemd-udevd"],
            ["systemctl", "start", "systemd-udevd"],
            ["systemctl", "start", "snapd"],
            ["getent", "hosts", "snapcraft.io"],
        ]

        check_call_calls = [
            args[0][len(LXD_RUN_COMMAND_PREFIX) :]
            for args, kwargs in self.check_call_mock.call_args_list
        ]

        assert check_call_calls == [
            [
                "mv",
                "/var/tmp/L2V0Yy9zeXN0ZW1kL25ldHdvcmsvMTAtZXRoMC5uZXR3b3Jr",
                "/etc/systemd/network/10-eth0.network",
            ],
            ["chown", "root:root", "/etc/systemd/network/10-eth0.network"],
            ["chmod", "0644", "/etc/systemd/network/10-eth0.network"],
            ["mv", "/var/tmp/L2V0Yy9ob3N0bmFtZQ==", "/etc/hostname"],
            ["chown", "root:root", "/etc/hostname"],
            ["chmod", "0644", "/etc/hostname"],
            ["apt-get", "update"],
            ["apt-get", "install", "dirmngr", "udev", "fuse", "--yes"],
            ["apt-get", "install", "snapd", "sudo", "--yes"],
            ["apt-get", "update"],
            ["apt-get", "dist-upgrade", "--yes"],
            ["apt-get", "install", "--yes", "apt-transport-https"],
            ["snap", "unset", "system", "proxy.http"],
            ["snap", "unset", "system", "proxy.https"],
        ]

    def test_clean_project(self):
        instance = LXDTestImpl(project=self.project, echoer=self.echoer_mock)
        instance.create()

        instance.clean_project()

        container = self.fake_pylxd_client.containers.get(self.instance_name)
        container.stop_mock.assert_called_once_with(wait=True)
        container.delete_mock.assert_called_once_with(wait=True)
        self.assertThat(instance.clean_project(), Equals(True))

    def test_clean_project_new_instance(self):
        pre_instance = LXDTestImpl(project=self.project, echoer=self.echoer_mock)
        pre_instance.create()
        pre_instance._stop()

        self.fake_pylxd_client.containers.get(self.instance_name)._reset_mocks()

        instance = LXDTestImpl(project=self.project, echoer=self.echoer_mock)
        instance.clean_project()

        container = self.fake_pylxd_client.containers.get(self.instance_name)
        container.stop_mock.assert_not_called()
        container.delete_mock.assert_called_once_with(wait=True)
        self.assertThat(instance.clean_project(), Equals(True))

    def test_clean_when_missing(self):
        instance = LXDTestImpl(project=self.project, echoer=self.echoer_mock)
        self.fake_pylxd_client.containers.get_mock.side_effect = pylxd.exceptions.NotFound(
            "not found"
        )

        instance.clean_project()

    def test_destroy_when_not_created(self):
        instance = LXDTestImpl(project=self.project, echoer=self.echoer_mock)
        # This call should not fail
        instance.destroy()

    def test_create_for_type_base(self):
        self.project._snap_meta.name = "core18"
        self.project._snap_meta.type = "base"
        self.project._snap_meta.base = None

        instance = LXDTestImpl(project=self.project, echoer=self.echoer_mock)

        instance.create()

        self.fake_pylxd_client.containers.create_mock.assert_called_once_with(
            config={
                "name": "snapcraft-core18",
                "raw.idmap": f"both {os.getuid()} 0",
                "security.syscalls.intercept.mknod": "true",
                "source": {
                    "mode": "pull",
                    "type": "image",
                    "server": "https://cloud-images.ubuntu.com/buildd/releases",
                    "protocol": "simplestreams",
                    "alias": "18.04",
                },
            },
            wait=True,
        )

    def test_create_invalid_base(self):
        self.project._snap_meta.base = "core19"

        instance = LXDTestImpl(project=self.project, echoer=self.echoer_mock)

        self.assertRaises(errors.ProviderInvalidBaseError, instance.create)

    def test_create_without_syscall_intercept(self):
        self.fake_pylxd_client.host_info["environment"]["kernel_features"][
            "seccomp_listener"
        ] = "false"

        instance = LXDTestImpl(project=self.project, echoer=self.echoer_mock)

        instance.create()

        self.fake_pylxd_client.containers.create_mock.assert_called_once_with(
            config={
                "name": "snapcraft-project-name",
                "raw.idmap": f"both {os.getuid()} 0",
                "source": {
                    "mode": "pull",
                    "type": "image",
                    "server": "https://cloud-images.ubuntu.com/buildd/releases",
                    "protocol": "simplestreams",
                    "alias": "16.04",
                },
            },
            wait=True,
        )


class LXDLaunchedTest(LXDBaseTest):
    def setUp(self):
        super().setUp()

        self.instance = LXDTestImpl(project=self.project, echoer=self.echoer_mock)
        self.instance.create()
        self.fake_container = self.fake_pylxd_client.containers.get(self.instance_name)
        # reset for the tests to only be concerned about what they are testing
        self.fake_container._reset_mocks()
        self.fake_pylxd_client.containers.create_mock.reset_mock()
        self.check_call_mock.reset_mock()

    def test_destroy(self):
        self.instance.destroy()

        self.fake_container.stop_mock.assert_called_once_with(wait=True)

    def test_destroy_and_launch(self):
        self.instance.destroy()

        self.fake_container.stop_mock.assert_called_once_with(wait=True)

        self.instance.create()

        self.fake_pylxd_client.containers.create_mock.assert_not_called()
        self.fake_container.start_mock.assert_called_once_with(wait=True)

    def test_pull_file(self):
        self.instance.pull_file("src.txt", "dest.txt")

        self.fake_container.files_get_mock.assert_called_once_with(file_name="src.txt")
        self.fake_container.files_delete_mock.assert_not_called()
        self.assertThat("dest.txt", FileExists())
        self.assertThat("dest.txt", FileContains("fake-pull"))

    def test_pull_file_and_delete(self):
        self.instance.pull_file("src.txt", "dest.txt", delete=True)

        self.fake_container.files_get_mock.assert_called_once_with(file_name="src.txt")
        self.fake_container.files_delete_mock.assert_called_once_with(
            file_name="src.txt"
        )
        self.assertThat("dest.txt", FileExists())
        self.assertThat("dest.txt", FileContains("fake-pull"))

    def test_push_file(self):
        with open("src.txt", "w") as f:
            f.write("fake-put")

        self.instance._push_file(source="src.txt", destination="dest.txt")

        self.fake_container.files_put_mock.assert_called_once_with(
            destination="dest.txt", contents=b"fake-put"
        )

    def test_shell(self):
        self.instance.shell()

        self.check_call_mock.assert_called_once_with(
            [*LXD_RUN_COMMAND_PREFIX, "/bin/bash"]
        )

    def test_mount_project(self):
        # project is mounted as dependency of create(), called in setUp().
        self.assertThat(
            self.fake_container.devices,
            Equals(
                {
                    "snapcraft-project": dict(
                        path="/root/project", source=self.path, type="disk"
                    )
                }
            ),
        )

    def test_mount_prime_directory(self):
        self.check_output_mock.return_value = b"/root"

        self.instance._mount_prime_directory()

        assert self.fake_container.devices.get("snapcraft-project-prime") == dict(
            path="/root/prime", source=os.path.join(self.path, "prime"), type="disk",
        )
        self.assertThat(self.fake_container.sync_mock.call_count, Equals(1))
        self.fake_container.save_mock.assert_called_once_with(wait=True)

    def test_run(self):
        self.instance._run(["ls", "/root/project"])

        self.check_call_mock.assert_called_once_with(
            [*LXD_RUN_COMMAND_PREFIX, "ls", "/root/project"]
        )


class EnsureLXDTest(LXDBaseTest):
    def test_linux(self):
        self.fake_snapd.snaps_result = [
            dict(name="lxd", channel="stable", revision="10")
        ]

        # Thou shall not fail
        with mock.patch(
            "snapcraft.internal.repo.Repo.is_package_installed", return_value=False
        ):
            LXD.ensure_provider()

    def test_linux_with_snap_and_deb_installed(self):
        self.fake_snapd.snaps_result = [
            dict(name="lxd", channel="stable", revision="10")
        ]

        # Thou shall not fail
        with mock.patch(
            "snapcraft.internal.repo.Repo.is_package_installed", return_value=True
        ):
            raised = self.assertRaises(SnapcraftEnvironmentError, LXD.ensure_provider)

        self.assertThat(
            str(raised),
            Equals(
                "The 'LXD' provider does not support having the 'lxd' "
                "or 'lxd-client' deb packages installed. "
                "To completely migrate to the LXD snap run 'lxd.migrate' "
                "and try again."
            ),
        )

    def test_lxd_snap_not_installed(self):
        raised = self.assertRaises(errors.ProviderNotFound, LXD.ensure_provider)

        self.assertThat(raised.prompt_installable, Equals(True))
        self.assertThat(
            raised.error_message,
            Equals("The LXD snap is required to continue: snap install lxd"),
        )

    def test_snap_support_missing(self):
        with mock.patch(
            "snapcraft.internal.repo.snaps.SnapPackage.is_snap_installed",
            side_effect=SnapdConnectionError(snap_name="lxd", url="fake"),
        ):
            raised = self.assertRaises(errors.ProviderNotFound, LXD.ensure_provider)

        self.assertThat(raised.prompt_installable, Equals(False))
        self.assertThat(
            raised.error_message,
            Equals(
                "snap support is required to continue: "
                "https://docs.snapcraft.io/installing-snapd/6735"
            ),
        )

    def test_non_linux(self):
        with mock.patch("sys.platform", return_value="darwin"):
            raised = self.assertRaises(errors.ProviderNotFound, LXD.ensure_provider)

        self.assertThat(raised.prompt_installable, Equals(False))
        self.assertThat(
            raised.error_message, Equals("LXD is not supported on this platform")
        )


class SetupLXDTest(LXDBaseTest):
    def setUp(self):
        super().setUp()

        self.fake_snapd.snaps_result = [
            dict(name="lxd", channel="stable", revision="10")
        ]
        self.fake_snapd.find_result = [
            dict(
                lxd=dict(
                    channel="stable",
                    type="app",
                    channels={"latest/stable": dict(confinement="strict")},
                )
            )
        ]

    def test_install(self):
        LXD.setup_provider(echoer=self.echoer_mock)

        self.assertThat(self.check_output_mock.call_count, Equals(2))
        self.check_output_mock.assert_has_calls(
            [
                mock.call(["/snap/bin/lxd", "waitready", "--timeout=30"]),
                mock.call(["/snap/bin/lxd", "init", "--auto"]),
            ]
        )

    def test_lxd_wait_fails(self):
        self.check_output_mock.side_effect = subprocess.CalledProcessError(
            cmd=["wait-fake"], returncode=1
        )

        raised = self.assertRaises(
            SnapcraftEnvironmentError, LXD.setup_provider, echoer=self.echoer_mock
        )

        self.assertThat(
            str(raised), Equals("Timeout reached waiting for LXD to start.")
        )

    def test_lxd_init_fails(self):
        self.check_output_mock.side_effect = [
            b"",
            subprocess.CalledProcessError(cmd=["init-fake"], returncode=1),
        ]

        raised = self.assertRaises(
            SnapcraftEnvironmentError, LXD.setup_provider, echoer=self.echoer_mock
        )

        self.assertThat(
            str(raised),
            Equals(
                "Failed to initialize LXD. "
                "Try manually initializing before trying again: lxd init --auto."
            ),
        )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2018 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import testscenarios

from tests import fixture_setup, integration


class UnicodePropertyTestCase(testscenarios.WithScenarios, integration.TestCase):

    scenarios = [
        ("summary", dict(name="foo", summary="bar💩", description="baz")),
        ("description", dict(name="foo", summary="bar", description="baz💩")),
    ]

    def test_invalid_unicode_workaround(self):
        snapcraft_yaml = fixture_setup.SnapcraftYaml(
            self.path,
            name=self.name,
            summary=self.summary,
            description=self.description,
        )
        snapcraft_yaml.update_part("my-part", {"plugin": "nil"})
        self.useFixture(snapcraft_yaml)
        self.run_snapcraft("pull")



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2016-2019 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

import re
from unittest import mock

import fixtures
import pytest
from testtools.matchers import Contains, Equals, MatchesRegex, Not

from snapcraft import storeapi

from . import FakeStoreCommandsBaseTestCase


class ExportLoginCommandTestCase(FakeStoreCommandsBaseTestCase):
    def setUp(self):
        super().setUp()

        self.useFixture(
            fixtures.MockPatchObject(
                storeapi.StoreClient,
                "acl",
                return_value={
                    "snap_ids": None,
                    "channels": None,
                    "permissions": None,
                    "expires": "2018-01-01T00:00:00",
                },
            )
        )

    def test_successful_export(self):
        result = self.run_command(
            ["export-login", "exported"], input="user@example.com\nsecret\n"
        )

        self.assertThat(result.exit_code, Equals(0))
        self.assertThat(result.output, Contains(storeapi.constants.TWO_FACTOR_WARNING))
        self.assertThat(result.output, Contains("Login successfully exported"))
        self.assertThat(
            result.output, MatchesRegex(r".*snaps:.*?No restriction", re.DOTALL)
        )
        self.assertThat(
            result.output, MatchesRegex(r".*channels:.*?No restriction", re.DOTALL)
        )
        self.assertThat(
            result.output, MatchesRegex(r".*permissions:.*?No restriction", re.DOTALL)
        )
        self.assertThat(
            result.output, MatchesRegex(r".*expires:.*?2018-01-01T00:00:00", re.DOTALL)
        )

        self.fake_store_login.mock.assert_called_once_with(
            email="user@example.com",
            password="secret",
            acls=None,
            packages=None,
            channels=None,
            expires=None,
            save=False,
            config_fd=None,
        )

    def test_successful_export_stdout(self):
        result = self.run_command(
            ["export-login", "-"], input="user@example.com\nsecret\n"
        )

        self.assertThat(result.exit_code, Equals(0))
        self.assertThat(result.output, Contains(storeapi.constants.TWO_FACTOR_WARNING))
        self.assertThat(result.output, Contains("Exported login starts on next line"))
        self.assertThat(
            result.output, Contains("Login successfully exported and printed above")
        )
        self.assertThat(
            result.output, MatchesRegex(r".*snaps:.*?No restriction", re.DOTALL)
        )
        self.assertThat(
            result.output, MatchesRegex(r".*channels:.*?No restriction", re.DOTALL)
        )
        self.assertThat(
            result.output, MatchesRegex(r".*permissions:.*?No restriction", re.DOTALL)
        )
        self.assertThat(
            result.output, MatchesRegex(r".*expires:.*?2018-01-01T00:00:00", re.DOTALL)
        )

        self.fake_store_login.mock.assert_called_once_with(
            email="user@example.com",
            password="secret",
            acls=None,
            packages=None,
            channels=None,
            expires=None,
            save=False,
            config_fd=None,
        )

    def test_successful_export_expires(self):
        self.useFixture(
            fixtures.MockPatchObject(
                storeapi.StoreClient,
                "acl",
                return_value={
                    "snap_ids": None,
                    "channels": None,
                    "permissions": None,
                    "expires": "2018-02-01T00:00:00",
                },
            )
        )

        result = self.run_command(
            ["export-login", "--expires=2018-02-01T00:00:00", "exported"],
            input="user@example.com\nsecret\n",
        )
        self.assertThat(result.exit_code, Equals(0))
        self.assertThat(result.output, Contains(storeapi.constants.TWO_FACTOR_WARNING))
        self.assertThat(result.output, Contains("Login successfully exported"))
        self.assertThat(
            result.output, MatchesRegex(r".*snaps:.*?No restriction", re.DOTALL)
        )
        self.assertThat(
            result.output, MatchesRegex(r".*channels:.*?No restriction", re.DOTALL)
        )
        self.assertThat(
            result.output, MatchesRegex(r".*permissions:.*?No restriction", re.DOTALL)
        )
        self.assertThat(
            result.output, MatchesRegex(r".*expires:.*?2018-02-01T00:00:00", re.DOTALL)
        )

        self.fake_store_login.mock.assert_called_once_with(
            email="user@example.com",
            password=mock.ANY,
            acls=None,
            packages=None,
            channels=None,
            expires="2018-02-01T00:00:00",
            save=False,
            config_fd=None,
        )

    def test_successful_login_with_2fa(self):
        self.fake_store_login.mock.side_effect = [
            storeapi.http_clients.errors.StoreTwoFactorAuthenticationRequired(),
            None,
        ]

        result = self.run_command(
            ["export-login", "exported"], input="user@example.com\nsecret\n123456"
        )

        self.assertThat(result.exit_code, Equals(0))
        self.assertThat(
            result.output, Not(Contains(storeapi.constants.TWO_FACTOR_WARNING))
        )
        self.assertThat(result.output, Contains("Login successfully exported"))
        self.assertThat(
            result.output, MatchesRegex(r".*snaps:.*?No restriction", re.DOTALL)
        )
        self.assertThat(
            result.output, MatchesRegex(r".*channels:.*?['edge']", re.DOTALL)
        )
        self.assertThat(
            result.output, MatchesRegex(r".*permissions:.*?No restriction", re.DOTALL)
        )
        self.assertThat(
            result.output, MatchesRegex(r".*expires:.*?2018-01-01T00:00:00", re.DOTALL)
        )

        self.assertThat(self.fake_store_login.mock.call_count, Equals(2))
        self.fake_store_login.mock.assert_has_calls(
            [
                mock.call(
                    email="user@example.com",
                    password="secret",
                    acls=None,
                    packages=None,
                    channels=None,
                    expires=None,
                    save=False,
                    config_fd=None,
                ),
                mock.call(
                    email="user@example.com",
                    password="secret",
                    otp="123456",
                    acls=None,
                    packages=None,
                    channels=None,
                    expires=None,
                    save=False,
                    config_fd=None,
                ),
            ]
        )

    def test_failed_login_with_invalid_credentials(self):
        self.fake_store_login.mock.side_effect = storeapi.http_clients.errors.InvalidCredentialsError(
            "error"
        )

        with pytest.raises(
            storeapi.http_clients.errors.InvalidCredentialsError
        ) as exc_info:
            self.run_command(
                ["export-login", "exported"],
                input="bad-user@example.com\nbad-password\n",
            )

        assert (
            str(exc_info.value)
            == 'Invalid credentials: error. Have you run "snapcraft login"?'
        )



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2015-2017 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.
import os
from textwrap import dedent

from testtools.matchers import Contains, Equals, FileContains

import snapcraft.internal.errors

from . import CommandBaseTestCase


class InitCommandTestCase(CommandBaseTestCase):
    def test_init_must_write_snapcraft_yaml(self):
        expected_yaml = dedent(
            """\
            name: my-snap-name # you probably want to 'snapcraft register <name>'
            base: core18 # the base snap is the execution environment for this snap
            version: '0.1' # just for humans, typically '1.2+git' or '1.3.2'
            summary: Single-line elevator pitch for your amazing snap # 79 char long summary
            description: |
              This is my-snap's description. You have a paragraph or two to tell the
              most important story about your snap. Keep it under 100 words though,
              we live in tweetspace and your description wants to look good in the snap
              store.

            grade: devel # must be 'stable' to release into candidate/stable channels
            confinement: devmode # use 'strict' once you have the right plugs and slots

            parts:
              my-part:
                # See 'snapcraft plugins'
                plugin: nil\n"""
        )

        result = self.run_command(["init"])

        self.assertThat(
            result.output,
            Contains(
                "Created snap/snapcraft.yaml.\n"
                "Go to https://docs.snapcraft.io/the-snapcraft-format/8337 for more "
                "information about the snapcraft.yaml format."
            ),
        )

        # Verify the generated yaml
        self.assertThat(
            os.path.join("snap", "snapcraft.yaml"), FileContains(expected_yaml)
        )


class InitCommandExistingProjectTestCase(CommandBaseTestCase):
    def test_init_with_existing_yaml(self):
        """Test that init bails if project yaml already exists"""
        yaml_path = os.path.join("snap", "snapcraft.yaml")

        dirname = os.path.dirname(yaml_path)
        if dirname:
            os.mkdir(dirname)

        open(yaml_path, "w").close()

        raised = self.assertRaises(
            snapcraft.internal.errors.SnapcraftEnvironmentError,
            self.run_command,
            ["init"],
        )

        self.assertThat(str(raised), Equals("{} already exists!".format(yaml_path)))



# -*- Mode:Python; indent-tabs-mode:nil; tab-width:4 -*-
#
# Copyright (C) 2020 Canonical Ltd
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License version 3 as
# published by the Free Software Foundation.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

from testtools import TestCase
from testtools.matchers import Equals

from snapcraft.plugins.v2.go import GoPlugin


class GoPluginTest(TestCase):
    def test_schema(self):
        schema = GoPlugin.get_schema()

        self.assertThat(
            schema,
            Equals(
                {
                    "$schema": "http://json-schema.org/draft-04/schema#",
                    "type": "object",
                    "additionalProperties": False,
                    "properties": {
                        "go-channel": {"type": "string", "default": "latest/stable"},
                        "go-buildtags": {
                            "type": "array",
                            "uniqueItems": True,
                            "items": {"type": "string"},
                            "default": [],
                        },
                    },
                    "required": ["source"],
                }
            ),
        )

    def test_get_build_snaps(self):
        class Options:
            go_channel = "14/latest"

        plugin = GoPlugin(part_name="my-part", options=Options())

        self.assertThat(plugin.get_build_snaps(), Equals({"go/14/latest"}))

    def test_get_build_packages(self):
        plugin = GoPlugin(part_name="my-part", options=lambda: None)

        self.assertThat(plugin.get_build_packages(), Equals({"gcc"}))

    def test_get_build_environment(self):
        plugin = GoPlugin(part_name="my-part", options=lambda: None)

        self.assertThat(
            plugin.get_build_environment(),
            Equals(
                {
                    "CGO_ENABLED": "1",
                    "GOBIN": "${SNAPCRAFT_PART_INSTALL}/bin",
                    "SNAPCRAFT_GO_LDFLAGS": "-ldflags -linkmode=external",
                }
            ),
        )

    def test_get_build_commands(self):
        class Options:
            go_buildtags = list()

        plugin = GoPlugin(part_name="my-part", options=Options())

        self.assertThat(
            plugin.get_build_commands(),
            Equals(
                [
                    "go mod download",
                    'go install -p "${SNAPCRAFT_PARALLEL_BUILD_COUNT}"  ${SNAPCRAFT_GO_LDFLAGS} ./...',
                ]
            ),
        )

    def test_get_build_commands_with_buildtags(self):
        class Options:
            go_buildtags = ["dev", "debug"]

        plugin = GoPlugin(part_name="my-part", options=Options())

        self.assertThat(
            plugin.get_build_commands(),
            Equals(
                [
                    "go mod download",
                    'go install -p "${SNAPCRAFT_PARALLEL_BUILD_COUNT}" -tags=dev,debug ${SNAPCRAFT_GO_LDFLAGS} ./...',
                ]
            ),
        )
